/**
  @file eve.c
 */
/*
 * ============================================================================
 * History
 * =======
 * 2017-03-15 : Created
 *
 * (C) Copyright Bridgetek Pte Ltd
 * ============================================================================
 *
 * This source code ("the Software") is provided by Bridgetek Pte Ltd
 * ("Bridgetek") subject to the licence terms set out
 * http://www.ftdichip.com/FTSourceCodeLicenceTerms.htm ("the Licence Terms").
 * You must read the Licence Terms before downloading or using the Software.
 * By installing or using the Software you agree to the Licence Terms. If you
 * do not agree to the Licence Terms then do not download or use the Software.
 *
 * Without prejudice to the Licence Terms, here is a summary of some of the key
 * terms of the Licence Terms (and in the event of any conflict between this
 * summary and the Licence Terms then the text of the Licence Terms will
 * prevail).
 *
 * The Software is provided "as is".
 * There are no warranties (or similar) in relation to the quality of the
 * Software. You use it at your own risk.
 * The Software should not be used in, or for, any medical device, system or
 * appliance. There are exclusions of Bridgetek liability for certain types of loss
 * such as: special loss or damage; incidental loss or damage; indirect or
 * consequential loss or damage; loss of income; loss of business; loss of
 * profits; loss of revenue; loss of contracts; business interruption; loss of
 * the use of money or anticipated savings; loss of information; loss of
 * opportunity; loss of goodwill or reputation; and/or loss of, damage to or
 * corruption of data.
 * There is a monetary cap on Bridgetek's liability.
 * The Software may have subsequently been amended by another user and then
 * distributed by that other user ("Adapted Software").  If so that user may
 * have additional licence terms that apply to those amendments. However, Bridgetek
 * has no liability in relation to those amendments.
 * ============================================================================
 */


/* INCLUDES ************************************************************************/

#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>


#include "FT_Platform.h"
#include "FT_Gpu.h"

#include <ft900_usb_hid.h>
#include <ft900_dlog.h>

#include "eve.h"
#include "tinyprintf.h"

extern volatile uint8_t D2XXTEST__Testmode;

/**
 @brief Link to datalogger area defined in crt0.S file.
 @details Must be passed to dlog library functions to initialise and use
        datalogger functions. We use the datalogger area for persistent
        configuration storage.
 */
extern __flash__ uint32_t __dlog_partition[];

/* CONSTANTS ***********************************************************************/

/**
 @brief Page number in datalogger memory in Flash for touchscreen calibration
 values.
 */
#define CONFIG_PAGE_TOUCHSCREEN 0
/**
 @brief Key for identifying if touchscreen calibration values are programmed into
 datalogger memory in the Flash.
 */
#define VALID_KEY_TOUCHSCREEN 0xd72f91a3

/**
 * @brief Show the built-in logo animation on the FT8xx.
 */
#undef SHOW_LOGO

/**
 * @brief Buffer display list and co-processor commands before sending them to
 * the FT8xx.
 * @details This will improve the speed at which commands are sent to the display
 * list or co-processor on the FT8xx. Commands are added to a buffer in RAM before
 * being written out as a block transaction over SPI when the command list is
 * completed. To start a command block use:
 * 	Ft_Gpu_CoCmd_Dlstart(phost);
 * When the command block is filled and ready to send then the following command
 * will send the buffer to the FT8xx and execute the list:
 * 	Ft_App_Flush_Co_Buffer(phost);
 * The program can wait until the list has been completed in the usual way with:
 * 	Ft_Gpu_Hal_WaitCmdfifo_empty(phost);
 * The size of the buffer can be adjusted.
 */
#undef FT8xx_BUFFER_COMMANDS

/**
 @brief Initial bootup code fro FT8xx.
 */
static const uint8_t FT_DLCODE_BOOTUP[12] =
{
		255, 255, 255, 2,//GPU instruction CLEAR_COLOR_RGB
		7, 0, 0, 38, //GPU instruction CLEAR
		0, 0, 0, 0,  //GPU instruction DISPLAY
};

/**
 @brief Custom font and bitmap definitions.
 @details These utilise handles 0 to 14.
 */
//@{
#define CUSTOM_BITMAP0 5
#define CUSTOM_BITMAP1 6
#define CUSTOM_BITMAP2 7
//@}

/* GLOBAL VARIABLES ****************************************************************/

/**
 @brief Pointers to custom images in external C file.
 */
//@{
extern const uint32_t img_bridgetek_icon_size;
extern const uint8_t __flash__ *img_bridgetek_icon_data;
/// Address in RAM_G
uint32_t img_bridgetek_icon_address;
uint32_t img_bridgetek_icon_width;
uint32_t img_bridgetek_icon_height;

extern const uint32_t img_bridgetek_logo_size;
extern const uint8_t __flash__ *img_bridgetek_logo_data;
/// Address in RAM_G
uint32_t img_bridgetek_logo_address;
uint32_t img_bridgetek_logo_width;
uint32_t img_bridgetek_logo_height;

uint32_t img_end_address;
//@}

/* LOCAL VARIABLES *****************************************************************/

/**
 @brief Area reserved to store configuration data which will be updated in
        datalogger storage area when required.
 */
static uint8_t	flashbuf[260] __attribute__((aligned(4)));

/**
 @brief Structure to hold touchscreen calibration settings.
 @details This is used to store the touchscreen calibration settings persistently
 in Flash and identify if the calibration needs to be re-performed.
 */
struct touchscreen_calibration {
	uint32_t key; // VALID_KEY_TOUCHSCREEN
	uint32_t transform[6];
};

/**
 * @brief Variables for display resolution to support various display panel
 * sizes and features.
 */
//@{
int16_t FT_DispWidth;
int16_t FT_DispHeight;
int16_t FT_DispHCycle;
int16_t FT_DispHOffset;
int16_t FT_DispHSync0;
int16_t FT_DispHSync1;
int16_t FT_DispVCycle;
int16_t FT_DispVOffset;
int16_t FT_DispVSync0;
int16_t FT_DispVSync1;
uint8_t FT_DispPCLK;
int8_t FT_DispSwizzle;
int8_t FT_DispPCLKPol;
int8_t FT_DispCSpread;
int8_t FT_DispDither;
//@}

/**
 * @brief Handle to the FT8xx
 */
Ft_Gpu_Hal_Context_t host, *phost;

uint32_t Ft_CmdBuffer_Index;
uint32_t Ft_DlBuffer_Index;

#ifdef FT8xx_BUFFER_COMMANDS
uint8_t  Ft_DlBuffer[FT_DL_SIZE];
uint8_t  Ft_CmdBuffer[FT_CMD_FIFO_SIZE];
#endif

#ifdef SHOW_LOGO
/**
 * @brief Information string for the splash screen.
 * @details this is 4 lines long and each line is added by a Ft_Gpu_CoCmd_Text.
 */
char *info[] = { "FT8xx/FT9xx Active Image Display",
		"BRT_AN_xxx",
		"Application Note to demonstrate interactive display device.",
		"Image sent via D2XX and rendered on display."
};
#endif // SHOW_LOGO


/* MACROS **************************************************************************/

/* LOCAL FUNCTIONS / INLINES *******************************************************/

/**
 * @brief Functions used to store calibration data in flash.
 */
//@{
int8_t flash_calib_init(void)
{
	int	pgsz;
	int	i;
	int ret;

	ret = dlog_init(__dlog_partition, &pgsz, &i);
	if (ret < 0)
	{
		// Project settings incorrect. Require dlog support with modified
		// linker script and crt0.S file.
		// See AN_398 for examples.
		return -1;
	}
	return 0;
}

int8_t flash_calib_write(struct touchscreen_calibration *calib)
{
	calib->key = VALID_KEY_TOUCHSCREEN;
	dlog_erase();
	memset(flashbuf, 0xff, sizeof(flashbuf));
	memcpy(flashbuf, calib, sizeof(struct touchscreen_calibration));
	if (dlog_prog(CONFIG_PAGE_TOUCHSCREEN, (uint32_t *)flashbuf) < 0)
	{
		// Flash not written.
		return -1;
	}
	return 0;
}

int8_t flash_calib_read(struct touchscreen_calibration *calib)
{
	memset(flashbuf, 0x00, sizeof(flashbuf));
	if (dlog_read(CONFIG_PAGE_TOUCHSCREEN, (uint32_t *)flashbuf) < 0)
	{
		return -1;
	}

	if (((struct touchscreen_calibration *)flashbuf)->key == VALID_KEY_TOUCHSCREEN)
	{
		memcpy(calib, flashbuf, sizeof(struct touchscreen_calibration));
		return 0;
	}
	return -2;
}
//@}

/**
 * @brief Write co-processor commands to the command buffer.
 */
void Ft_App_WrCoCmd_Buffer(Ft_Gpu_Hal_Context_t *phost, uint32_t cmd)
{
#ifdef FT8xx_BUFFER_COMMANDS

	/* Copy the command instruction into buffer */
	uint32_t *pBuffcmd;
	pBuffcmd =(uint32_t*)&Ft_CmdBuffer[Ft_CmdBuffer_Index];
	*pBuffcmd = cmd;
	/* Increment the command index */
	Ft_CmdBuffer_Index += FT_CMD_SIZE;

#else // FT8xx_BUFFER_COMMANDS

	Ft_Gpu_Hal_WrCmd32(phost, cmd);

#endif
}

void Ft_App_WrDlCmd_Buffer(Ft_Gpu_Hal_Context_t *phost, uint32_t cmd)
{
#ifdef FT8xx_BUFFER_COMMANDS

	/* Copy the command instruction into buffer */
	uint32_t *pBuffcmd;
	pBuffcmd =(uint32_t*)&Ft_DlBuffer[Ft_DlBuffer_Index];
	*pBuffcmd = cmd;

#else // FT8xx_BUFFER_COMMANDS

	Ft_Gpu_Hal_Wr32(phost, (RAM_DL+Ft_DlBuffer_Index), cmd);

#endif
	/* Increment the command index */
	Ft_DlBuffer_Index += FT_CMD_SIZE;
}

void Ft_App_WrCoStr_Buffer(Ft_Gpu_Hal_Context_t *phost, const int8_t *s)
{
#ifdef  FT8xx_BUFFER_COMMANDS

	uint16_t length = 0;
	length = strlen(s) + 1;//last for the null termination

	strcpy(&Ft_CmdBuffer[Ft_CmdBuffer_Index], s);

	/* increment the length and align it by 4 bytes */
	Ft_CmdBuffer_Index += ((length + 3) & ~3);

#endif
}

void Ft_App_Flush_DL_Buffer(Ft_Gpu_Hal_Context_t *phost)
{
#ifdef  FT8xx_BUFFER_COMMANDS
	if (Ft_DlBuffer_Index> 0)
		Ft_Gpu_Hal_WrMem(phost, RAM_DL, Ft_DlBuffer, Ft_DlBuffer_Index);
#endif
	Ft_DlBuffer_Index = 0;

}

void Ft_App_Flush_Co_Buffer(Ft_Gpu_Hal_Context_t *phost)
{
#ifdef  FT8xx_BUFFER_COMMANDS
	if (Ft_CmdBuffer_Index > 0)
		Ft_Gpu_Hal_WrCmdBuf(phost, Ft_CmdBuffer, Ft_CmdBuffer_Index);
#endif
	Ft_CmdBuffer_Index = 0;
}


void Ft_BootupConfig()
{
	uint8_t chipid;

	Ft_Gpu_Hal_Powercycle(phost, FT_TRUE);

	/* Access address 0 to wake up the FT800 */
	Ft_Gpu_HostCommand(phost, FT_GPU_ACTIVE_M);
	Ft_Gpu_Hal_Sleep(20);

	/* Set the clk to external clock */
#if (FT800_PLATFORM != ME800A_HV35R) && (FT800_PLATFORM != ME810A_HV35R) \
		&& (FT800_PLATFORM != ME812A_WH50R)
	Ft_Gpu_HostCommand(phost, FT_GPU_EXTERNAL_OSC);
	Ft_Gpu_Hal_Sleep(10);
#endif

	//Read Register ID to check if FT800 is ready.
	chipid = Ft_Gpu_Hal_Rd8(phost, REG_ID);
	while(chipid != 0x7C)
	{
		chipid = Ft_Gpu_Hal_Rd8(phost, REG_ID);
		ft_delay(100);
	}
	Ft_Gpu_81X_SelectSysCLK(phost, FT_GPU_SYSCLK_60M);

	/* Configuration of LCD display */
#if DISPLAY_RESOLUTION ==  DISPLAY_RESOLUTION_QVGA
	/* Values specific to QVGA LCD display */
	FT_DispWidth = 320;
	FT_DispHeight = 240;
	FT_DispHCycle =  408;
	FT_DispHOffset = 70;
	FT_DispHSync0 = 0;
	FT_DispHSync1 = 10;
	FT_DispVCycle = 263;
	FT_DispVOffset = 13;
	FT_DispVSync0 = 0;
	FT_DispVSync1 = 2;
	FT_DispPCLK = 8;
	FT_DispSwizzle = 2;
	FT_DispPCLKPol = 0;
	FT_DispCSpread = 1;
	FT_DispDither = 1;

#elif DISPLAY_RESOLUTION == DISPLAY_RESOLUTION_WVGA
	/* Values specific to QVGA LCD display */
	FT_DispWidth = 800;
	FT_DispHeight = 480;
	FT_DispHCycle =  928;
	FT_DispHOffset = 88;
	FT_DispHSync0 = 0;
	FT_DispHSync1 = 48;
	FT_DispVCycle = 525;
	FT_DispVOffset = 32;
	FT_DispVSync0 = 0;
	FT_DispVSync1 = 3;
	FT_DispPCLK = 2;
	FT_DispSwizzle = 0;
	FT_DispPCLKPol = 1;
	FT_DispCSpread = 0;
	FT_DispDither = 1;
#elif DISPLAY_RESOLUTION == DISPLAY_RESOLUTION_HVGA_PORTRAIT
	/* Values specific to HVGA LCD display */

	FT_DispWidth = 320;
	FT_DispHeight = 480;
	FT_DispHCycle =  400;
	FT_DispHOffset = 40;
	FT_DispHSync0 = 0;
	FT_DispHSync1 = 10;
	FT_DispVCycle = 500;
	FT_DispVOffset = 10;
	FT_DispVSync0 = 0;
	FT_DispVSync1 = 5;
	FT_DispPCLK = 4;
	FT_DispSwizzle = 2;
	FT_DispPCLKPol = 1;
	FT_DispCSpread = 1;
	FT_DispDither = 1;

#ifdef ME810A_HV35R
	FT_DispPCLK = 5;
#endif

#endif // DISPLAY_RESOLUTION

#if (FT800_PLATFORM == ME800A_HV35R) || (FT800_PLATFORM == ME810A_HV35R) \
		|| (FT800_PLATFORM == ME812A_WH50R)
	/* After recognizing the type of chip, perform the trimming if necessary */
	Ft_Gpu_ClockTrimming(phost, LOW_FREQ_BOUND);
#endif

	Ft_Gpu_Hal_Wr16(phost, REG_HCYCLE, FT_DispHCycle);
	Ft_Gpu_Hal_Wr16(phost, REG_HOFFSET, FT_DispHOffset);
	Ft_Gpu_Hal_Wr16(phost, REG_HSYNC0, FT_DispHSync0);
	Ft_Gpu_Hal_Wr16(phost, REG_HSYNC1, FT_DispHSync1);
	Ft_Gpu_Hal_Wr16(phost, REG_VCYCLE, FT_DispVCycle);
	Ft_Gpu_Hal_Wr16(phost, REG_VOFFSET, FT_DispVOffset);
	Ft_Gpu_Hal_Wr16(phost, REG_VSYNC0, FT_DispVSync0);
	Ft_Gpu_Hal_Wr16(phost, REG_VSYNC1, FT_DispVSync1);
	Ft_Gpu_Hal_Wr8(phost, REG_SWIZZLE, FT_DispSwizzle);
	Ft_Gpu_Hal_Wr8(phost, REG_PCLK_POL, FT_DispPCLKPol);
	Ft_Gpu_Hal_Wr16(phost, REG_HSIZE, FT_DispWidth);
	Ft_Gpu_Hal_Wr16(phost, REG_VSIZE, FT_DispHeight);
	Ft_Gpu_Hal_Wr16(phost, REG_CSPREAD, FT_DispCSpread);
	Ft_Gpu_Hal_Wr16(phost, REG_DITHER, FT_DispDither);

#if (FT800_DEVICE == FT_800_ENABLE) || (FT800_DEVICE == FT_810_ENABLE) \
		|| (FT800_DEVICE == FT_812_ENABLE)
	/* Touch configuration - configure the resistance value to 1200 - this value is specific to customer requirement and derived by experiment */
	Ft_Gpu_Hal_Wr16(phost, REG_TOUCH_RZTHRESH, RESISTANCE_THRESHOLD);
#endif

	Ft_Gpu_Hal_Wr8(phost, REG_GPIO_DIR, 0xff);
	Ft_Gpu_Hal_Wr8(phost, REG_GPIO, 0xff);

	/*It is optional to clear the screen here*/
	Ft_Gpu_Hal_WrMem(phost, RAM_DL, (uint8_t *)FT_DLCODE_BOOTUP, sizeof(FT_DLCODE_BOOTUP));
	Ft_Gpu_Hal_Wr8(phost, REG_DLSWAP, DLSWAP_FRAME);

	Ft_Gpu_Hal_Wr8(phost, REG_PCLK, FT_DispPCLK);//after this display is visible on the LCD


#ifdef ENABLE_ILI9488_HVGA_PORTRAIT
	/* to cross check reset pin */
	Ft_Gpu_Hal_Wr8(phost, REG_GPIO, 0xff);
	ft_delay(120);
	Ft_Gpu_Hal_Wr8(phost, REG_GPIO, 0x7f);
	ft_delay(120);
	Ft_Gpu_Hal_Wr8(phost, REG_GPIO, 0xff);

	ILI9488_Bootup();

	/* Reconfigure the SPI */
#ifdef FT900_PLATFORM
	printf("after ILI9488 bootup \n");
	//spi
	// Initialize SPIM HW
	sys_enable(sys_device_spi_master);
	gpio_function(27, pad_spim_sck); /* GPIO27 to SPIM_CLK */
	gpio_function(28, pad_spim_ss0); /* GPIO28 as CS */
	gpio_function(29, pad_spim_mosi); /* GPIO29 to SPIM_MOSI */
	gpio_function(30, pad_spim_miso); /* GPIO30 to SPIM_MISO */

	gpio_write(28, 1);
	spi_init(SPIM, spi_dir_master, spi_mode_0, 4);
#endif

#endif

	/* make the spi to quad mode - addition 2 bytes for silicon */
#ifdef FT_81X_ENABLE
	/* api to set quad and numbe of dummy bytes */
#if SPI_ENABLE == ENABLE_SPI_QUAD
	Ft_Gpu_Hal_SetSPI(phost, FT_GPU_SPI_QUAD_CHANNEL, FT_GPU_SPI_TWODUMMY);
#elif SPI_ENABLE ==  ENABLE_SPI_DUAL
	Ft_Gpu_Hal_SetSPI(phost, FT_GPU_SPI_DUAL_CHANNEL, FT_GPU_SPI_TWODUMMY);
#elif SPI_ENABLE == ENABLE_SPI_SINGLE
	Ft_Gpu_Hal_SetSPI(phost, FT_GPU_SPI_SINGLE_CHANNEL, FT_GPU_SPI_ONEDUMMY);
#endif

#endif

#ifdef FT900_PLATFORM
	spi_init(SPIM, spi_dir_master, spi_mode_0, 16);

#if SPI_ENABLE == ENABLE_SPI_QUAD
	/* Initialize IO2 and IO3 pad/pin for dual and quad settings */
	gpio_function(31, pad_spim_io2);
	gpio_function(32, pad_spim_io3);
	gpio_write(31, 1);
	gpio_write(32, 1);
#endif

	spi_option(SPIM, spi_option_fifo_size, 64);
	spi_option(SPIM, spi_option_fifo, 1);
	spi_option(SPIM, spi_option_fifo_receive_trigger, 1);
#endif

#if SPI_ENABLE == ENABLE_SPI_QUAD
#ifdef FT900_PLATFORM
	spi_option(SPIM, spi_option_bus_width, 4);
#endif
#elif SPI_ENABLE == ENABLE_SPI_DUAL
#ifdef FT900_PLATFORM
	spi_option(SPIM, spi_option_bus_width, 2);
#endif
#elif SPI_ENABLE == ENABLE_SPI_SINGLE
#ifdef FT900_PLATFORM
	spi_option(SPIM, spi_option_bus_width, 1);
#endif


#endif

	phost->ft_cmd_fifo_wp = Ft_Gpu_Hal_Rd16(phost, REG_CMD_WRITE);
}

uint8_t Read_Keypad(uint8_t *key)
{
	uint8_t Read_tag;
	uint8_t key_detect = 0;

	Read_tag = Ft_Gpu_Hal_Rd8(phost, REG_TOUCH_TAG);

	if (!(Ft_Gpu_Hal_Rd16(phost, REG_TOUCH_RAW_XY) & 0x8000))
	{
		key_detect = 1;
		*key = Read_tag;
	}

	return key_detect;
}

uint8_t Read_Touchscreen(uint16_t *x, uint16_t *y)
{
	uint8_t key_detect = 0;
	uint32_t touch;

	touch = Ft_Gpu_Hal_Rd32(phost, REG_TOUCH_SCREEN_XY);

	if (touch & 0x8000)
	{
		// no touch
		key_detect = 0;
	}
	else
	{
		key_detect = 1;
		*x = (touch >> 16) & 0x7fff;
		*y = touch & 0x7fff;
	}

	return key_detect;
}

void Play_Sound(uint8_t sound, uint8_t volume)
{
	Ft_Gpu_Hal_Wr8(phost, REG_VOL_SOUND, volume);
	Ft_Gpu_Hal_Wr8(phost, REG_SOUND, sound);
	Ft_Gpu_Hal_Wr8(phost, REG_PLAY, 1);
}

void eve_calibrate()
{
	struct touchscreen_calibration calib;
	Ft_CmdBuffer_Index = 0;

	if (flash_calib_read(&calib) != 0)
	{
		Ft_Gpu_CoCmd_Dlstart(phost);
		Ft_App_WrCoCmd_Buffer(phost, CLEAR_COLOR_RGB(0, 0, 0));
		Ft_App_WrCoCmd_Buffer(phost, CLEAR(1, 1, 1));
		Ft_App_WrCoCmd_Buffer(phost, COLOR_RGB(255, 255, 255));
		Ft_Gpu_CoCmd_Text(phost, FT_DispWidth/2, FT_DispHeight/2,
				27, OPT_CENTERX|OPT_CENTERY,"Please tap on a dot");
		Ft_Gpu_CoCmd_Calibrate(phost, 0);
		Ft_App_Flush_Co_Buffer(phost);
		Ft_Gpu_Hal_WaitCmdfifo_empty(phost);

		calib.transform[0] = Ft_Gpu_Hal_Rd32(phost, REG_TOUCH_TRANSFORM_A);
		calib.transform[1] = Ft_Gpu_Hal_Rd32(phost, REG_TOUCH_TRANSFORM_B);
		calib.transform[2] = Ft_Gpu_Hal_Rd32(phost, REG_TOUCH_TRANSFORM_C);
		calib.transform[3] = Ft_Gpu_Hal_Rd32(phost, REG_TOUCH_TRANSFORM_D);
		calib.transform[4] = Ft_Gpu_Hal_Rd32(phost, REG_TOUCH_TRANSFORM_E);
		calib.transform[5] = Ft_Gpu_Hal_Rd32(phost, REG_TOUCH_TRANSFORM_F);
		flash_calib_write(&calib);
	}
	Ft_Gpu_Hal_Wr32(phost, REG_TOUCH_TRANSFORM_A, calib.transform[0]);
	Ft_Gpu_Hal_Wr32(phost, REG_TOUCH_TRANSFORM_B, calib.transform[1]);
	Ft_Gpu_Hal_Wr32(phost, REG_TOUCH_TRANSFORM_C, calib.transform[2]);
	Ft_Gpu_Hal_Wr32(phost, REG_TOUCH_TRANSFORM_D, calib.transform[3]);
	Ft_Gpu_Hal_Wr32(phost, REG_TOUCH_TRANSFORM_E, calib.transform[4]);
	Ft_Gpu_Hal_Wr32(phost, REG_TOUCH_TRANSFORM_F, calib.transform[5]);
}

void eve_splash()
{
#ifdef SHOW_LOGO
	uint16_t dloffset = 0, z;
	uint8_t sk;

	Ft_CmdBuffer_Index = 0;

	Ft_Gpu_CoCmd_Dlstart(phost);
	Ft_App_WrCoCmd_Buffer(phost, CLEAR_COLOR_RGB(0, 0, 0));
	Ft_App_WrCoCmd_Buffer(phost, CLEAR(1, 1, 1));
	Ft_App_WrCoCmd_Buffer(phost, COLOR_RGB(255, 255, 255));
	Ft_App_WrCoCmd_Buffer(phost, DISPLAY());
	Ft_Gpu_CoCmd_Swap(phost);
	Ft_App_Flush_Co_Buffer(phost);
	Ft_Gpu_Hal_WaitCmdfifo_empty(phost);

	Ft_Gpu_CoCmd_Logo(phost);
	Ft_App_Flush_Co_Buffer(phost);
	Ft_Gpu_Hal_WaitCmdfifo_empty(phost);
	while(0!=Ft_Gpu_Hal_Rd16(phost, REG_CMD_READ));

	dloffset = Ft_Gpu_Hal_Rd16(phost, REG_CMD_DL);
	dloffset -= 4;

#ifdef FT_81X_ENABLE
	dloffset -= 2*4;//remove two more instructions in case of 81x
#endif // FT_81X_ENABLE
	Ft_Gpu_Hal_WrCmd32(phost, CMD_MEMCPY);
	Ft_Gpu_Hal_WrCmd32(phost, 100000L);
	Ft_Gpu_Hal_WrCmd32(phost, RAM_DL);
	Ft_Gpu_Hal_WrCmd32(phost, dloffset);

	do
	{
		Ft_Gpu_CoCmd_Dlstart(phost);
		Ft_Gpu_CoCmd_Append(phost, 100000L, dloffset);
		Ft_App_WrCoCmd_Buffer(phost, BITMAP_TRANSFORM_A(256));
		Ft_App_WrCoCmd_Buffer(phost, BITMAP_TRANSFORM_A(256));
		Ft_App_WrCoCmd_Buffer(phost, BITMAP_TRANSFORM_B(0));
		Ft_App_WrCoCmd_Buffer(phost, BITMAP_TRANSFORM_C(0));
		Ft_App_WrCoCmd_Buffer(phost, BITMAP_TRANSFORM_D(0));
		Ft_App_WrCoCmd_Buffer(phost, BITMAP_TRANSFORM_E(256));
		Ft_App_WrCoCmd_Buffer(phost, BITMAP_TRANSFORM_F(0));
		Ft_App_WrCoCmd_Buffer(phost, SAVE_CONTEXT());
		Ft_App_WrCoCmd_Buffer(phost, COLOR_RGB(219, 180, 150));
		Ft_App_WrCoCmd_Buffer(phost, COLOR_A(220));
		Ft_App_WrCoCmd_Buffer(phost, BEGIN(EDGE_STRIP_A));
		Ft_App_WrCoCmd_Buffer(phost, VERTEX2F(0, FT_DispHeight*16));
		Ft_App_WrCoCmd_Buffer(phost, VERTEX2F(FT_DispWidth*16, FT_DispHeight*16));
		Ft_App_WrCoCmd_Buffer(phost, COLOR_A(255));
		Ft_App_WrCoCmd_Buffer(phost, RESTORE_CONTEXT());
		Ft_App_WrCoCmd_Buffer(phost, COLOR_RGB(0, 0, 0));
		// INFORMATION
		Ft_Gpu_CoCmd_Text(phost, FT_DispWidth/2, 20, 28, OPT_CENTERX|OPT_CENTERY, info[0]);
		Ft_Gpu_CoCmd_Text(phost, FT_DispWidth/2, 60, 26, OPT_CENTERX|OPT_CENTERY, info[1]);
		Ft_Gpu_CoCmd_Text(phost, FT_DispWidth/2, 90, 26, OPT_CENTERX|OPT_CENTERY, info[2]);
		Ft_Gpu_CoCmd_Text(phost, FT_DispWidth/2, 120, 26, OPT_CENTERX|OPT_CENTERY, info[3]);
		Ft_Gpu_CoCmd_Text(phost, FT_DispWidth/2, FT_DispHeight-30, 26, OPT_CENTERX|OPT_CENTERY,"Click to start");
		Read_Keypad(&sk);
		if(sk!='P')
			Ft_App_WrCoCmd_Buffer(phost, COLOR_RGB(255, 255, 255));
		else
			Ft_App_WrCoCmd_Buffer(phost, COLOR_RGB(100, 100, 100));
		Ft_App_WrCoCmd_Buffer(phost, BEGIN(FTPOINTS));
		Ft_App_WrCoCmd_Buffer(phost, POINT_SIZE(20*16));
		Ft_App_WrCoCmd_Buffer(phost, TAG('P'));
		Ft_App_WrCoCmd_Buffer(phost, VERTEX2F((FT_DispWidth/2)*16, (FT_DispHeight-60)*16));
		Ft_App_WrCoCmd_Buffer(phost, COLOR_RGB(180, 35, 35));
		Ft_App_WrCoCmd_Buffer(phost, BEGIN(BITMAPS));
		Ft_App_WrCoCmd_Buffer(phost, VERTEX2II((FT_DispWidth/2)-14, (FT_DispHeight-75), 14, 4));
		Ft_App_WrCoCmd_Buffer(phost, DISPLAY());
		Ft_Gpu_CoCmd_Swap(phost);
		Ft_App_Flush_Co_Buffer(phost);
		Ft_Gpu_Hal_WaitCmdfifo_empty(phost);
	}
	while(sk != 'P');
	Play_Sound(0x51, 255);
#endif // SHOW_LOGO
}

/* FUNCTIONS ***********************************************************************/

void Ft_App_Get_Props(Ft_Gpu_Hal_Context_t *phost, uint32_t *addr,
		uint32_t *width, uint32_t *height)
{
	uint32_t copro_write;

	Ft_Gpu_CoCmd_Dlstart(phost);
	// To read the result from CMD_GETPROPS we need to be clever and find out
	// there the CoProcessor is writing the command. We can then retrieve the
	// results from the place where they were written.
	copro_write = Ft_Gpu_Hal_Rd32(phost, REG_CMD_WRITE);
	// Send the command to the CoProcessor.
	Ft_Gpu_CoCmd_GetProps(phost, 0, 0, 0);
	// Make it happen.
	Ft_App_WrCoCmd_Buffer(phost, DISPLAY());
	// Wait for it to finish.
	Ft_App_Flush_Co_Buffer(phost);
	Ft_Gpu_Hal_WaitCmdfifo_empty(phost);

	// Obtain the results from the RAM_CMD in the CoProcessor.
	Ft_Gpu_Hal_RdMem(phost, RAM_CMD + copro_write + sizeof(uint32_t), (uint8_t *)addr, sizeof(uint32_t));
	Ft_Gpu_Hal_RdMem(phost, RAM_CMD + copro_write + 2 * sizeof(uint32_t), (uint8_t *)width, sizeof(uint32_t));
	Ft_Gpu_Hal_RdMem(phost, RAM_CMD + copro_write + 3 * sizeof(uint32_t), (uint8_t *)height, sizeof(uint32_t));
}

void eve_load_images()
{
	img_bridgetek_icon_address = 0;

	Ft_Gpu_CoCmd_Dlstart(phost);
	Ft_App_WrCoCmd_Buffer(phost, BITMAP_HANDLE(CUSTOM_BITMAP0));
	Ft_Gpu_CoCmd_LoadImage(phost, img_bridgetek_icon_address, 0);
	Ft_Gpu_Hal_WrCmdBufFromFlash(phost, img_bridgetek_icon_data, img_bridgetek_icon_size);
	Ft_Gpu_CoCmd_Swap(phost);
	Ft_App_Flush_Co_Buffer(phost);
	Ft_Gpu_Hal_WaitCmdfifo_empty(phost);

	Ft_App_Get_Props(phost, &img_bridgetek_icon_address,
			&img_bridgetek_icon_width, &img_bridgetek_icon_height);
	img_bridgetek_logo_address = img_bridgetek_icon_address +
			(img_bridgetek_icon_width * 2 * img_bridgetek_icon_height);
	img_bridgetek_logo_address += 3;
	img_bridgetek_logo_address &= (~3);

	Ft_Gpu_CoCmd_Dlstart(phost);
	Ft_App_WrCoCmd_Buffer(phost, BITMAP_HANDLE(CUSTOM_BITMAP1));
	Ft_Gpu_CoCmd_LoadImage(phost, img_bridgetek_logo_address, 0); // to be above both fonts above.
	Ft_Gpu_Hal_WrCmdBufFromFlash(phost, img_bridgetek_logo_data, img_bridgetek_logo_size);
	Ft_Gpu_CoCmd_Swap(phost);
	Ft_App_Flush_Co_Buffer(phost);
	Ft_Gpu_Hal_WaitCmdfifo_empty(phost);

	Ft_App_Get_Props(phost, &img_bridgetek_logo_address,
			&img_bridgetek_logo_width, &img_bridgetek_logo_height);

	img_end_address = img_bridgetek_logo_address +
			(img_bridgetek_logo_width * 2 * img_bridgetek_logo_height);
	img_end_address += 3;
	img_end_address &= (~3);
}

void eve_setup()
{
	Ft_Gpu_HalInit_t halinit;

	flash_calib_init();

	ft_millis_init();

	halinit.TotalChannelNum = 1;

	// FT9xx specific setup of SPI Master
	Ft_Gpu_Hal_Init(&halinit);
	host.hal_config.channel_no = 0;
	host.hal_config.pdn_pin_no = FT800_PD_N;
	host.hal_config.spi_cs_pin_no = FT800_SEL_PIN;
	Ft_Gpu_Hal_Open(&host);

	phost = &host;

	Ft_BootupConfig();
	Ft_Gpu_Hal_WrMem(phost, RAM_DL, (uint8_t *)FT_DLCODE_BOOTUP, sizeof(FT_DLCODE_BOOTUP));

	Ft_Gpu_Hal_Wr8(phost, REG_DLSWAP, DLSWAP_FRAME);

	Ft_Gpu_Hal_Sleep(200);

	eve_calibrate();
#ifdef SHOW_LOGO
	eve_splash();
#endif // SHOW_LOGO

	eve_load_images();

	Ft_Gpu_CoCmd_Dlstart(phost);
	Ft_App_WrCoCmd_Buffer(phost, CLEAR_COLOR_RGB(0, 0, 0));
	Ft_App_WrCoCmd_Buffer(phost, CLEAR(1, 1, 1));
	Ft_App_WrCoCmd_Buffer(phost, COLOR_RGB(255, 255, 255));
	Ft_App_WrCoCmd_Buffer(phost, DISPLAY());
	Ft_Gpu_CoCmd_Swap(phost);
	Ft_App_Flush_Co_Buffer(phost);
	Ft_Gpu_Hal_WaitCmdfifo_empty(phost);
}

void eve_wait(void)
{
	Ft_Gpu_CoCmd_Dlstart(phost);
	Ft_App_WrCoCmd_Buffer(phost, CLEAR_COLOR_RGB(0, 0, 0));
	Ft_App_WrCoCmd_Buffer(phost, CLEAR(1, 1, 1));
	Ft_App_WrCoCmd_Buffer(phost, COLOR_RGB(255, 255, 255));
	Ft_Gpu_CoCmd_Text(phost, FT_DispWidth/2, FT_DispHeight/2,
			27, OPT_CENTERX|OPT_CENTERY,"Waiting for host...");

	Ft_App_WrCoCmd_Buffer(phost, BITMAP_HANDLE(CUSTOM_BITMAP1));
	Ft_App_WrCoCmd_Buffer(phost, BEGIN(BITMAPS));
	Ft_App_WrCoCmd_Buffer(phost, BITMAP_SOURCE(img_bridgetek_logo_address));
	Ft_App_WrCoCmd_Buffer(phost, BITMAP_LAYOUT(RGB565, img_bridgetek_logo_width*2, img_bridgetek_logo_height));
	Ft_App_WrCoCmd_Buffer(phost, BITMAP_SIZE(NEAREST, BORDER, BORDER,
			img_bridgetek_logo_width, img_bridgetek_logo_height));
	Ft_App_WrCoCmd_Buffer(phost, BITMAP_LAYOUT_H((img_bridgetek_logo_width * 2) >> 10, img_bridgetek_logo_height >> 9));
	Ft_App_WrCoCmd_Buffer(phost, BITMAP_SIZE_H(img_bridgetek_logo_width >> 9, img_bridgetek_logo_height >> 9));
	Ft_App_WrCoCmd_Buffer(phost, VERTEX2II((FT_DispWidth/2)-(img_bridgetek_logo_width/2), 0, CUSTOM_BITMAP1, 0));
	Ft_App_WrCoCmd_Buffer(phost, DISPLAY());
	Ft_Gpu_CoCmd_Swap(phost);
	Ft_App_Flush_Co_Buffer(phost);
	Ft_Gpu_Hal_WaitCmdfifo_empty(phost);

	delayms(100);
}

void eve_start(void)
{
	Ft_Gpu_CoCmd_Dlstart(phost);
	Ft_App_WrCoCmd_Buffer(phost, CLEAR_COLOR_RGB(0, 0, 0));
	Ft_App_WrCoCmd_Buffer(phost, CLEAR(1, 1, 1));
	Ft_App_WrCoCmd_Buffer(phost, COLOR_RGB(255, 255, 255));
	Ft_Gpu_CoCmd_Text(phost, FT_DispWidth/2, FT_DispHeight/2,
			27, OPT_CENTERX|OPT_CENTERY,"Waiting for images...");

	Ft_App_WrCoCmd_Buffer(phost, BITMAP_HANDLE(CUSTOM_BITMAP1));
	Ft_App_WrCoCmd_Buffer(phost, BEGIN(BITMAPS));
	Ft_App_WrCoCmd_Buffer(phost, BITMAP_SOURCE(img_bridgetek_logo_address));
	Ft_App_WrCoCmd_Buffer(phost, BITMAP_LAYOUT(RGB565, img_bridgetek_logo_width*2, img_bridgetek_logo_height));
	Ft_App_WrCoCmd_Buffer(phost, BITMAP_SIZE(NEAREST, BORDER, BORDER,
			img_bridgetek_logo_width, img_bridgetek_logo_height));
	Ft_App_WrCoCmd_Buffer(phost, BITMAP_LAYOUT_H((img_bridgetek_logo_width * 2) >> 10, img_bridgetek_logo_height >> 9));
	Ft_App_WrCoCmd_Buffer(phost, BITMAP_SIZE_H(img_bridgetek_logo_width >> 9, img_bridgetek_logo_height >> 9));
	Ft_App_WrCoCmd_Buffer(phost, VERTEX2II((FT_DispWidth/2)-(img_bridgetek_logo_width/2), 0, CUSTOM_BITMAP1, 0));
	Ft_App_WrCoCmd_Buffer(phost, DISPLAY());
	Ft_Gpu_CoCmd_Swap(phost);
	Ft_App_Flush_Co_Buffer(phost);
	Ft_Gpu_Hal_WaitCmdfifo_empty(phost);
}

int8_t eve_out_loop(uint8_t *buffer, int32_t *length)
{
	static int8_t pressed = 0;

	uint8_t touch;
	uint16_t touch_x;
	uint16_t touch_y;

	touch = Read_Touchscreen(&touch_x, &touch_y);

	if (!touch)
	{
		// no touch
		pressed = 0;
	}
	else
	{
		if (!pressed)
		{
			*length = snprintf((char *)buffer, *length, "%d %d\r\n", touch_x, touch_y);
			pressed = 1;
		}
		else
		{
			touch = 0;
		}
	}

	return touch;
}

int8_t eve_in_loop(uint8_t *buffer, int32_t length)
{
	static uint32_t image_start_address;
	static uint32_t image_len = 0;
	static uint8_t eof_pending = 0;
	static uint8_t frame_active = 0;

	uint32_t image_addr;
	uint32_t image_width;
	uint32_t image_height;
	uint8_t *pmark;
	uint8_t *pstart = buffer;

	if (!frame_active)
	{
		if ((buffer[0] == 0xff) && (buffer[1] == 0xd8))
		{
			// Start of JPEG marker found.
			image_start_address = 0;
			image_len = 0;
			frame_active = 1;

			Ft_Gpu_CoCmd_Dlstart(phost);
			Ft_App_WrCoCmd_Buffer(phost, BITMAP_HANDLE(CUSTOM_BITMAP2));
			Ft_Gpu_CoCmd_LoadImage(phost, image_start_address, OPT_NODL);
			Ft_App_Flush_Co_Buffer(phost);
		}
	}

	if (frame_active)
	{
		image_len += length;
		length = (length + 3) & (~3);

		Ft_Gpu_Hal_WrCmdBuf_nowait(phost, buffer, length);

		// Find 0xff 0xd9 for end of image!
		if (eof_pending)
		{
			if (*pstart == 0xd9)
			{
				frame_active = 0;
			}
			eof_pending = 0;
		}

		pmark = pstart;
		while (pmark)
		{
			pmark = memchr(pmark, 0xff, length - (pmark - pstart));
			if (pmark)
			{
				pmark++;
				if ((pmark - pstart) == length)
				{
					// The last byte of the sample is 0xff so there might be
					// and end of frame marker continued on the next sample.
					eof_pending = 1;
				}
				else if (*pmark == 0xd9)
				{
					frame_active = 0;
				}
			}
		}

		if (!frame_active)
		{
			Ft_App_Get_Props(phost, &image_addr,
					&image_width, &image_height);

			Ft_Gpu_CoCmd_Dlstart(phost);
			Ft_App_WrCoCmd_Buffer(phost, CLEAR_COLOR_RGB(0, 0, 0));
			Ft_App_WrCoCmd_Buffer(phost, CLEAR(1, 1, 1));
			Ft_App_WrCoCmd_Buffer(phost, COLOR_RGB(255, 255, 255));
			Ft_App_WrCoCmd_Buffer(phost, BITMAP_HANDLE(CUSTOM_BITMAP2));
			Ft_App_WrCoCmd_Buffer(phost, BEGIN(BITMAPS));
			Ft_App_WrCoCmd_Buffer(phost, BITMAP_SOURCE(image_addr));
			Ft_App_WrCoCmd_Buffer(phost, BITMAP_LAYOUT(RGB565, image_width * 2, image_height));
			Ft_App_WrCoCmd_Buffer(phost, BITMAP_LAYOUT_H((image_width * 2) >> 10, image_height >> 9));
			Ft_App_WrCoCmd_Buffer(phost, BITMAP_SIZE(NEAREST, BORDER, BORDER,
					image_width, image_height));
			Ft_App_WrCoCmd_Buffer(phost, BITMAP_SIZE_H(image_width >> 9, image_height >> 9));
			Ft_App_WrCoCmd_Buffer(phost, VERTEX2II(
					(FT_DispWidth / 2) - (image_width / 2),
					(FT_DispHeight / 2) - (image_height / 2),
					CUSTOM_BITMAP2, 0));
			Ft_App_WrCoCmd_Buffer(phost, DISPLAY());
			Ft_Gpu_CoCmd_Swap(phost);
			Ft_App_Flush_Co_Buffer(phost);
			Ft_Gpu_Hal_WaitCmdfifo_empty(phost);
		}
	}

	return frame_active;
}

void eve_close(void)
{
	/* Close all the opened handles */
	Ft_Gpu_Hal_Close(phost);
	Ft_Gpu_Hal_DeInit();
}

void eve_tick(void)
{
	millis_ticker();
}
