/**
  @file main.c
  @brief
  D2XX project template.

 */
/*
 * ============================================================================
 * History
 * =======
 * 2017-03-22 : Created
 *
 * (C) Copyright Bridgetek Pte Ltd
 * ============================================================================
 *
 * This source code ("the Software") is provided by Bridgetek Pte Ltd
 *  ("Bridgetek ") subject to the licence terms set out
 * http://www.ftdichip.com/FTSourceCodeLicenceTerms.htm ("the Licence Terms").
 * You must read the Licence Terms before downloading or using the Software.
 * By installing or using the Software you agree to the Licence Terms. If you
 * do not agree to the Licence Terms then do not download or use the Software.
 *
 * Without prejudice to the Licence Terms, here is a summary of some of the key
 * terms of the Licence Terms (and in the event of any conflict between this
 * summary and the Licence Terms then the text of the Licence Terms will
 * prevail).
 *
 * The Software is provided "as is".
 * There are no warranties (or similar) in relation to the quality of the
 * Software. You use it at your own risk.
 * The Software should not be used in, or for, any medical device, system or
 * appliance. There are exclusions of Bridgetek liability for certain types of loss
 * such as: special loss or damage; incidental loss or damage; indirect or
 * consequential loss or damage; loss of income; loss of business; loss of
 * profits; loss of revenue; loss of contracts; business interruption; loss of
 * the use of money or anticipated savings; loss of information; loss of
 * opportunity; loss of goodwill or reputation; and/or loss of, damage to or
 * corruption of data.
 * There is a monetary cap on Bridgetek's liability.
 * The Software may have subsequently been amended by another user and then
 * distributed by that other user ("Adapted Software").  If so that user may
 * have additional licence terms that apply to those amendments. However, Bridgetek
 * has no liability in relation to those amendments.
 * ============================================================================
 */

/* INCLUDES ************************************************************************/

#include <stdint.h>
#include <string.h>
#include <ctype.h>

#include "ftd2xx_api.h"
#include <ft900_startup_dfu.h>
#include <ft900.h>

#include "FT_Platform.h"
#include "eve.h"

/* UART support for printf output. */
#include "tinyprintf.h"

#if defined(__GNUC__)
/* Soft ISR Vector table as defined in crt0.S */
extern isrptr_t vector_table[N_INTERRUPTS];
#elif defined(__MIKROC_PRO_FOR_FT90x__)
/* Soft ISR Vector table as defined in crt0.S */
extern isrptr_t vector_table[N_INTERRUPTS];
#endif
isrptr_t d2xx_timer_ISR;

#define BRIDGE_DEBUG
#ifdef BRIDGE_DEBUG
#define BRIDGE_DEBUG_PRINTF(...) do {tfp_printf(__VA_ARGS__);} while (0)
#else
#define BRIDGE_DEBUG_PRINTF(...)
#endif


/* MACROS ***********************************************************************/
#if defined(__FT930__)
/// FTDI Product ID for FT930 variants
#define USB_PID_FTDI_FT930_1                      0x6034
#define USB_PID_FTDI_FT930_2                      0x6035
#define USB_PID_FTDI_FT930_3                      0x6036
#define USB_PID_FTDI_FT930_4                      0x6037
#define USB_PID_FTDI_FT930_5                      0x6038
#define USB_PID_FTDI_FT930_6                      0x6039
#define USB_PID_FTDI_FT930_7                      0x603A
#else
/// FTDI Product ID for FT900 variants
#define USB_PID_FTDI_FT900_1                      0x6031
#define USB_PID_FTDI_FT900_2                      0x6032
#define USB_PID_FTDI_FT900_3                      0x6033
#endif

/**
 * @brief Enable/Disable and define the pin for remote wakeup.
 */
#undef GPIO_REMOTE_WAKEUP

#ifdef GPIO_REMOTE_WAKEUP
#if defined(__FT930__)
//GPIO12 in FT930Q mini module board
#define GPIO_PIN  12
#else
#define GPIO_PIN  18
#endif
#endif

/**
 * @brief Verify the checksum for the D2XX configuration file.
 * @details The checksum is stored in the ft900_d2xx_default_config.inc or
 * ft930_d2xx_default_config.inc file. This is loaded by crt0.S file.
 * The D2XX_Init function verifies the checksum and calls an error if it is
 * incorrect. When these default configuration files are modified then the
 * checksum needs to be calculated. This define will turn on code that
 * sends the required checksum to the UART if D2XX_Init fails.
 *
 * NOTE: if default configuration file is modified then the crt0.O in the
 * build directory MUST be deleted to force a rebuild of crt0.S with the
 * modified values.
 */
#define VERIFY_CHECKSUM

#define MAX_READBUFFER_SIZE  512

/* GLOBAL VARIABLES ****************************************************************/

extern TD2XX_DeviceConfiguration __pD2XXDefaultConfiguration[]; // pointer to D2XX Config in flash
// The test config (in Flash and then PM) is copied to the RAM structure D2XXTEST_FlashConfig on startup
__attribute__ ((aligned (4))) TD2XX_DeviceConfiguration D2XXTEST_UserD2xxConfig = {0};

/* LOCAL VARIABLES *****************************************************************/

/**
 @brief Millisecond counter
 @details Count-up timer to provide the elapsed time for network operations.
 */
static uint32_t milliseconds = 0;

static uint8_t D2XXTest_ReadBuffer[MAX_READBUFFER_SIZE + 16];
static char *D2XXTest_EventStrings[D2XX_EVT_MAX_CODE] = {	"SUSPEND",    	/**< SUSPEND EVENT from USB Host */
																	"RESUME",	 	/**< RESUME EVENT from USB Host */
																	"BUS_RESET",	 		/**< USB Bus Reset */
																	"READY",	 		/**< D2XX enters Ready state where READ/WRITE requests are processed*/
																	"UNREADY",	 	/**< D2XX exits Ready state*/
																	"DETACH",		/**< DFU DETACH Class command from DFU application */
																	"TESTMODE",	 	/**< D2XX enters Test Mode. Exit is via power cycle*/
																	"INTF_RESET"	/**< Interface RESET Vendor Command from D2XX Application */
																};
uint8_t D2XXTEST__DfuDetach = 0;
uint8_t D2XXTEST__Ready = 0;
volatile uint8_t SetRemoteWakeup = 0;
#if defined(__FT930__)
extern volatile uint8_t RemoteWakeupEnable;  // used in d2xx_api_master.c
#else
volatile uint8_t RemoteWakeupEnable = 0;
#endif
volatile uint8_t D2XXTEST__Suspend = 0;
volatile uint8_t D2XXTEST__Sleep = 0;
volatile uint8_t D2XXTEST__Wakeup = 0;
volatile uint8_t wkupSource = 0;
volatile uint8_t D2XXTEST__Testmode = 0;

/* LOCAL FUNCTIONS / INLINES *******************************************************/

void tfp_putc(void* p, char c);
void image_viewer_loop(void);
void d2xx_callback(ED2XX_EventCode  eventID, void *ref, void* param1, void* param2);
#ifdef GPIO_REMOTE_WAKEUP
void gpio_ISR();
#endif
#ifdef VERIFY_CHECKSUM
uint16_t xorChecksum(const char str[], uint16_t length);
#endif

/** @name tfp_putc
 *  @details Machine dependent putc function for tfp_printf (tinyprintf) library.
 *  @param p Parameters (machine dependent)
 *  @param c The character to write
 */
void tfp_putc(void* p, char c)
{
	uart_write((ft900_uart_regs_t*)p, (uint8_t)c);
}

/** @name d2xx_callback
 *  @details Callback function to the D2XX library.
 *  @param
 *  @param
 */
void d2xx_callback(ED2XX_EventCode  eventID, void *ref, void* param1, void* param2)
{
	/* 0 => Device, 1-3 => Interface*/
	/* or */
	/* 0 => Remote Wakeup disabled, 1 => Remote wakeup enabled */
	uint8_t param = 0;

	if (param1)
	{
		param = (*(uint8_t *)param1);
	}

#ifdef BRIDGE_DEBUG
	tfp_printf("~%d ",eventID);
	tfp_printf("%s\r\n", D2XXTest_EventStrings[eventID]);
#endif

	switch(eventID)
	{
	case D2XX_EVT_SUSPEND:
		if (D2XXTEST__Ready == 1)
		{
			D2XXTEST__Suspend = 1;
			D2XXTEST__Sleep = 0;
			RemoteWakeupEnable = param;
			tfp_printf("RemoteWakeup Enabled :%d \r\n", RemoteWakeupEnable);
		}
		break;
	case D2XX_EVT_RESUME:
	case D2XX_EVT_BUS_RESET:
		if (D2XXTEST__Sleep)
		{
			wkupSource |= 0x10;
			D2XXTEST__Wakeup = 1;
		}
		break;
	case D2XX_EVT_READY: /*Connected */
		D2XXTEST__Ready = 1;
		break;
	case D2XX_EVT_UNREADY: /*Detached */
		D2XXTEST__Ready = 0;
		D2XXTEST__Suspend = 1;
		break;
	case D2XX_EVT_DFU_DETACH: /* DFU Detach from DFU application */
		D2XXTEST__DfuDetach = 1;
		break;
	case D2XX_EVT_TESTMODE: /* Testmode during electrical tests */
		D2XXTEST__Testmode = 1;
		break;
	default:
		break;
	}
}

#ifdef VERIFY_CHECKSUM
/* function to calculate the expected XOR checksum for the d2xx configuration */
uint16_t xorChecksum(const char str[], uint16_t length) {
    uint16_t sum = 0;
    uint16_t *s = (uint16_t *)str;

    length >>= 1;
    while (length) {
        sum ^= *s++;
        length--;
    }
    return sum;
}
#endif // VERIFY_CHECKSUM

/** @brief Returns the current millisecond country
 *  @returns A count of milliseconds
 */
uint32_t millis(void)
{
	return milliseconds;
}

void timer_ISR(void)
{
	if (timer_is_interrupted(timer_select_a))
	{
		milliseconds++;
	}
	if (timer_is_interrupted(timer_select_b))
	{
		eve_tick();
	}

	d2xx_timer_ISR();
}

/* Power management ISR */
void power_management_ISR(void)
{

#if defined(__FT930__)
	if (SYS->PMCFG_H & MASK_SYS_PMCFG_SLAVE_PERI_IRQ_PEND)
	{
		// Clear d2xx hw engine wakeup
		SYS->PMCFG_H = MASK_SYS_PMCFG_SLAVE_PERI_IRQ_PEND;

		//D2XXTEST__Suspend = 0;
		if (D2XXTEST__Sleep)
		{
			wkupSource |= 1;
			D2XXTEST__Wakeup = 1;
		}
	}
#endif
	if (SYS->PMCFG_H & MASK_SYS_PMCFG_PM_GPIO_IRQ_PEND)
	{
		// Clear GPIO wakeup pending
		SYS->PMCFG_H = MASK_SYS_PMCFG_PM_GPIO_IRQ_PEND;
		if (D2XXTEST__Sleep)
		{
			wkupSource |= 2;
			D2XXTEST__Wakeup = 1;
		}
	}
#if defined(__FT900__)
	if (SYS->PMCFG_H & MASK_SYS_PMCFG_DEV_CONN_DEV)
	{
		// Clear connection interrupt
		SYS->PMCFG_H = MASK_SYS_PMCFG_DEV_CONN_DEV;
		if (D2XXTEST__Sleep)
		{
			wkupSource |= 4;
			D2XXTEST__Wakeup = 1;
		}

	}

	if (SYS->PMCFG_H & MASK_SYS_PMCFG_DEV_DIS_DEV)
	{
		// Clear disconnection interrupt
		SYS->PMCFG_H = MASK_SYS_PMCFG_DEV_DIS_DEV;
		if (D2XXTEST__Sleep)
		{
			wkupSource |= 8;
			D2XXTEST__Wakeup = 1;
		}
	}

	if (SYS->PMCFG_H & MASK_SYS_PMCFG_HOST_RST_DEV)
	{
		// Clear Host Reset interrupt
		SYS->PMCFG_H = MASK_SYS_PMCFG_HOST_RST_DEV;
		USBD_resume();
	}

	if (SYS->PMCFG_H & MASK_SYS_PMCFG_HOST_RESUME_DEV)
	{
		// Clear Host Resume interrupt
		SYS->PMCFG_H = MASK_SYS_PMCFG_HOST_RESUME_DEV;
		if(! (SYS->MSC0CFG & MASK_SYS_MSC0CFG_DEV_RMWAKEUP))
		{
			// If we are driving K-state on Device USB port;
			// We must maintain the 1ms requirement before resuming the phy
			USBD_resume();
		}
	}
#endif
}


#ifdef GPIO_REMOTE_WAKEUP
void gpio_ISR()
{
    if (gpio_is_interrupted(GPIO_PIN))
    {
		if (D2XXTEST__Sleep)
		{
			//Need to perform remote wakeup procedure
			D2XXTEST__Wakeup = 1;
			SetRemoteWakeup = 1;
		}
    }
}
#endif

void image_viewer_loop(void)
{
	ED2XX_ErrorCode retVal = 0;

	retVal = D2XX_Read(1, D2XXTest_ReadBuffer, MAX_READBUFFER_SIZE);
	if (retVal > 0)
	{
		eve_in_loop(D2XXTest_ReadBuffer, (int32_t)retVal);
	}
	else
	{
		// No data read from D2XX.
	}
}

void screen_touch_loop(void)
{
	ED2XX_ErrorCode retVal = 0;
	uint8_t buffer[16];
	int32_t buflen = sizeof(buffer);

	if (eve_out_loop(buffer, &buflen))
	{
		retVal = D2XX_Write(1, &buffer[0], buflen);
#ifdef BRIDGE_DEBUG
		tfp_printf("%s", buffer);
#endif
	}
}

uint8_t d2xx_testing(void)
{
	uint8_t not_connected = 1;

	// Setup the display.
	eve_setup();

	// Draw a waiting for host message.
	eve_wait();

	for (;;)
	{
    	/* Test process loop */
		if (D2XXTEST__Testmode)
		{
			tfp_printf("D2XXTEST__Testmode\r\n");
			for (;;);
			// wait for power cycle after testmode
		}
		// Application loop
		if (D2XXTEST__Ready)
		{
			if (not_connected)
			{
				// Now we are connected, draw the keyboard.
				tfp_printf("START\r\n");
				eve_start();
				not_connected = 0;
			}
			else
			{
				image_viewer_loop();
				screen_touch_loop();
			}
		}

		/* Event handling */
		if (D2XXTEST__DfuDetach)
		{
			tfp_printf("D2XXTEST__DfuDetach\r\n");
			delayms(100);
			/* DFU detach request had come. Exit D2XX */
			D2XX_Exit();

			not_connected = 1;

#if defined(__FT930__)
		    *(SLAVECPU) &= ~(MASK_SLAVE_CPU_CTRL_D2XX_MODE);    // turn-off D2XX_mode
		    *(SLAVECPU) |= (MASK_SLAVE_CPU_CTRL_SLV_RESET);  // assert bit to keep slave CPU in reset
#endif

			/* Take the system to DFU Mode. Provide timeout in ms to
			 * wait for re-enumeration with the host
			 * */
			//STARTUP_DFU(0);
			break;
		}

		/* Suspend mode handling */
		if (D2XXTEST__Suspend)
		{
			tfp_printf("D2XXTEST__Suspend\r\n");
			eve_wait();

			wkupSource = 0;
			tfp_printf("Entering Suspend Mode...\r\n");
			delayms(100);
#ifdef GPIO_REMOTE_WAKEUP
			//Configuring GPIO pin to wakeup
			SetRemoteWakeup = 0;
			if (RemoteWakeupEnable)
			{
				/* Set up the pin */
				gpio_dir(GPIO_PIN, pad_dir_input);
				gpio_pull(GPIO_PIN, pad_pull_pullup);

				/* Attach an interrupt */
				gpio_interrupt_enable(GPIO_PIN, gpio_int_edge_falling);
				interrupt_attach(interrupt_gpio, (uint8_t)interrupt_gpio, gpio_ISR);
			}
#endif

			//Enable d2xx hw engine wakeup events
#if defined(__FT930__)
			SYS->PMCFG_L &= ~(MASK_SYS_PMCFG_SLAVE_PERI_IRQ_EN);
			interrupt_attach(interrupt_0, (int8_t)interrupt_0, power_management_ISR);
			SYS->PMCFG_L |= MASK_SYS_PMCFG_SLAVE_PERI_IRQ_EN;
#endif

			// Close UART
#if 0
			uart_close(UART0);
			sys_disable(sys_device_uart0);
#endif


			/* Power down Mode */
			D2XXTEST__Sleep = 1;
			D2XXTEST__Wakeup = 0;
			SYS->PMCFG_L |= MASK_SYS_PMCFG_PM_PWRDN_MODE;
			SYS->PMCFG_L |= MASK_SYS_PMCFG_PM_PWRDN;

			asm volatile ("nop");
			asm volatile ("nop");
			asm volatile ("nop");
			asm volatile ("nop");
			asm volatile ("nop");

			//wait for wakeup
			while(!D2XXTEST__Wakeup);

			//Woken up!!! Disable all wakeup interrupt sources
			interrupt_disable_globally();
#if defined(__FT930__)
			interrupt_detach(interrupt_0);
			SYS->PMCFG_L &= ~(MASK_SYS_PMCFG_SLAVE_PERI_IRQ_EN);
#endif

			//Remove powerdown mode
			SYS->PMCFG_L &= ~MASK_SYS_PMCFG_PM_PWRDN_MODE;
			SYS->PMCFG_L &= ~MASK_SYS_PMCFG_PM_PWRDN;

#ifdef GPIO_REMOTE_WAKEUP
			gpio_interrupt_disable(GPIO_PIN);
			interrupt_detach(interrupt_gpio);
			gpio_is_interrupted(GPIO_PIN);
#endif

			D2XXTEST__Sleep = 0;
			D2XXTEST__Suspend = 0;
			D2XXTEST__Wakeup = 0;

			//Enable back the UART
#if 0
			debug_uart_init();
#endif
			interrupt_enable_globally();
			tfp_printf("Exiting Suspend Mode.. Wakeup Source: %02X \r\n", wkupSource);
			if (SetRemoteWakeup)
			{
				D2XX_IOCTL(0, D2XX_IOCTL_SYS_REMOTE_WAKEUP, &SetRemoteWakeup, NULL);
				SetRemoteWakeup = 0;
				tfp_printf("Sending remote wakeup to host.. \r\n");
			}

			eve_start();
		}
    }

	return 0;
}

/* FUNCTIONS ***********************************************************************/

int main(void)
{
	ED2XX_ErrorCode retVal;

	D2XXTEST__Suspend = 0;
	D2XXTEST__Sleep = 0;
	D2XXTEST__Wakeup = 0;

	sys_reset_all();
	sys_disable(sys_device_usb_device);

	/* Enable the UART Device... */
	sys_enable(sys_device_uart0);
#if defined(__FT900__)
    gpio_function(48, pad_uart0_txd); /* UART0 TXD */
    gpio_function(49, pad_uart0_rxd); /* UART0 RXD */
#elif defined(__FT930__)
    gpio_function(23, pad_uart0_txd); /* UART0 TXD */
    gpio_function(22, pad_uart0_rxd); /* UART0 RXD */
#endif

	/* Enable tfp_printf() functionality... */
	init_printf(UART0, tfp_putc);

	// Open the UART using the coding required.
	uart_open(UART0, 1, UART_DIVIDER_19200_BAUD, uart_data_bits_8, uart_parity_none, uart_stop_bits_1);

#ifdef BRIDGE_DEBUG
	/* Print out a welcome message... */
	BRIDGE_DEBUG_PRINTF("\x1B[2J" /* ANSI/VT100 - Clear the Screen */
			"\x1B[H"  /* ANSI/VT100 - Move Cursor to Home */
	);

	BRIDGE_DEBUG_PRINTF("(C) Copyright 2016, Bridgetek Pte Ltd. \r\n");
	BRIDGE_DEBUG_PRINTF("--------------------------------------------------------------------- \r\n");
	BRIDGE_DEBUG_PRINTF("Welcome to the USB Active Image Display... \r\n");
	BRIDGE_DEBUG_PRINTF("\r\n");
	BRIDGE_DEBUG_PRINTF("Emulate a USB Display Device with touchscreen feedback.\r\n");
	BRIDGE_DEBUG_PRINTF("--------------------------------------------------------------------- \r\n");
#endif // BRIDGE_DEBUG

	/* Attach the interrupt so it can be called... */
	//interrupt_attach(interrupt_uart0, (uint8_t) interrupt_uart0, uart0ISR);
	uart_disable_interrupt(UART0, uart_interrupt_tx);
	/* Enable the UART to fire interrupts when receiving buffer... */
	uart_enable_interrupt(UART0, uart_interrupt_rx);
	/* Enable interrupts to be fired... */
	uart_enable_interrupts_globally(UART0);

	sys_enable(sys_device_timer_wdt);
	/* Timer A = 1ms */
#if defined(__FT930__)
	timer_prescaler(timer_select_a, 1000);
#else
	timer_prescaler(1000);
#endif
	timer_init(timer_select_a, 100, timer_direction_down, timer_prescaler_select_on, timer_mode_continuous);
	timer_enable_interrupt(timer_select_a);
	timer_start(timer_select_a);

#if defined(__FT930__)
	retVal = D2XX_Init(&D2XXTEST_UserD2xxConfig, d2xx_callback, NULL);

	tfp_printf("D2XX_Init() returned: %d, No of Interfaces: %d\r\n", retVal,
			D2XXTEST_UserD2xxConfig.ConfigDesc.NumOfD2XXInterfaces);

    /*slave sub-system control register setup*/
    *(SLAVECPU) |= (MASK_SLAVE_CPU_CTRL_SLV_RESET);  // assert bit to keep slave CPU in reset
    *(SLAVECPU) |= (MASK_SLAVE_CPU_CTRL_D2XX_MODE);    // turn-on D2XX_mode
    *(SLAVECPU) &= ~(MASK_SLAVE_CPU_CTRL_SLV_RESET); // de-assert bit to allow slave CPU to start
#else
   	retVal = D2XX_Init(__pD2XXDefaultConfiguration, d2xx_callback, NULL);
	memcpy_pm2dat(&D2XXTEST_UserD2xxConfig.ConfigDesc,
			(uint32_t)&__pD2XXDefaultConfiguration->ConfigDesc,
			sizeof(TConfigDescriptors));

	tfp_printf("D2XX_Init() returned: %d, No of Interfaces: %d\r\n", retVal,
			D2XXTEST_UserD2xxConfig.ConfigDesc.NumOfD2XXInterfaces);
#endif

#ifdef VERIFY_CHECKSUM
	// Checksum may be wrong... give guidance.
	if (retVal == D2XX_ERR_INVALID_PARAMETER)
	{
		__attribute__ ((aligned (4))) TD2XX_DeviceConfiguration d2xx_device_config;
		memcpy_pm2dat((void *)&d2xx_device_config, (uint32_t)__pD2XXDefaultConfiguration, sizeof(TD2XX_DeviceConfiguration));

		tfp_printf("Configuration:- Signature: %08X, Actual Checksum: %04X, Expected Checksum : %04X\r\n",
				d2xx_device_config.Signature,
				d2xx_device_config.Checksum,
				xorChecksum((char *)&d2xx_device_config, sizeof(TD2XX_DeviceConfiguration) - 2));
		// NOTE: if ft9xx_d2xx_default_config.inc is modified then the crt0.O in the build
		// directory MUST be deleted to force a rebuild of crt0.S with the modified values.
		while (1);
	}
#endif // VERIFY_CHECKSUM

	d2xx_timer_ISR = vector_table[interrupt_timers];

	interrupt_attach(interrupt_timers, (int8_t)interrupt_timers, timer_ISR);

	/* Enable power management interrupts. Primarily to detect resume signalling
	 * from the USB host. */
#if defined(__FT900__)
	interrupt_attach(interrupt_0, (int8_t)interrupt_0, power_management_ISR);
#endif

	interrupt_enable_globally();

	d2xx_testing();

	interrupt_disable_globally();

	// Wait forever...
	for (;;);

	return 0;
}
