/**
 @file main.c
 @breif Main file for streaming data to/from a D2XX device.
 */
/*
 * ============================================================================
 * History
 * =======
 * 2017-03-15 : Created
 *
 * (C) Copyright Bridgetek Pte Ltd
 * ============================================================================
 *
 * This source code ("the Software") is provided by Bridgetek Pte Ltd
 * ("Bridgetek") subject to the licence terms set out
 * http://www.ftdichip.com/FTSourceCodeLicenceTerms.htm ("the Licence Terms").
 * You must read the Licence Terms before downloading or using the Software.
 * By installing or using the Software you agree to the Licence Terms. If you
 * do not agree to the Licence Terms then do not download or use the Software.
 *
 * Without prejudice to the Licence Terms, here is a summary of some of the key
 * terms of the Licence Terms (and in the event of any conflict between this
 * summary and the Licence Terms then the text of the Licence Terms will
 * prevail).
 *
 * The Software is provided "as is".
 * There are no warranties (or similar) in relation to the quality of the
 * Software. You use it at your own risk.
 * The Software should not be used in, or for, any medical device, system or
 * appliance. There are exclusions of Bridgetek liability for certain types of loss
 * such as: special loss or damage; incidental loss or damage; indirect or
 * consequential loss or damage; loss of income; loss of business; loss of
 * profits; loss of revenue; loss of contracts; business interruption; loss of
 * the use of money or anticipated savings; loss of information; loss of
 * opportunity; loss of goodwill or reputation; and/or loss of, damage to or
 * corruption of data.
 * There is a monetary cap on Bridgetek's liability.
 * The Software may have subsequently been amended by another user and then
 * distributed by that other user ("Adapted Software").  If so that user may
 * have additional licence terms that apply to those amendments. However, Bridgetek
 * has no liability in relation to those amendments.
 * ============================================================================
 */

/* INCLUDES ************************************************************************/

#include <stdio.h>
#include <malloc.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <time.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#elif defined _MSC_VER
// Building with Visual Studio
#include <windows.h>
#elif defined __WIN32
#include <conio.h>
#include <windows.h>
#include <direct.h>
#include <windef.h>
#include <winnt.h>
#include <winbase.h>
#include <process.h>
#include <fcntl.h>
#include <wctype.h>
#elif !defined __WIN32
#include <pthread.h>
#include <io.h>
#include <unistd.h>
#endif

#undef FTDI_EXPORTS
#include "ftd2xx.h"
#define FTDI_EXPORTS

/* CONSTANTS ***********************************************************************/

#ifdef FTDI_EXPORTS
#ifdef _WIN32
#define FTDI_API __declspec(dllexport)
#else
#define FTDI_API
#endif
#else
#define FTDI_API __declspec(dllimport)
#endif

/* Macro for calling convention, if required */
#ifdef _WIN32
//#define CAL_CONV
#define CAL_CONV	__stdcall
#else
#define CAL_CONV
#endif

#ifndef __WIN32
#define GET_FUNC(libHandle,symbol)	symbol
#else
#define GET_FUNC(libHandle,symbol) GetProcAddress(libHandle,symbol)
#define CHECK_SYMBOL(exp) CHECK_NULL(exp);
#endif

/* GLOBAL VARIABLES ****************************************************************/

/* LOCAL VARIABLES *****************************************************************/

HANDLE hdll_d2xx = NULL;
FT_HANDLE ftInterface;

/**
 @brief File/release version of utility.
 */
#define TTYUSB_VERSION "1.0.1"

#define TTYUSB_NORMAL 0
#define TTYUSB_ERROR 1
#define TTYUSB_EXIT_PARAMETERS -1
#define TTYUSB_EXIT_NO_D2XX -2

/* USB TYP port name for D2XX access */
char *ttyUSBPortName = NULL;
/* Input and output filenames */
char *inFilename = NULL;
char *outFilename = NULL;
/* Use stdin for input */
int useStdin = 0;
int useStdout = 0;
int useStdinout = 1;
int monitoring = 0;
/* Show logging information */
int logOutput = 0;
/* List all D2XX ports */
int listAll = 0;

// Windows...
HANDLE hStdin;

typedef FT_STATUS (CAL_CONV *pfunc_FT_GetLibraryVersion)(LPDWORD lpdwVersion);
typedef FT_STATUS (CAL_CONV *pfunc_FT_GetNumChannel)(LPDWORD lpdwNumDevs);
typedef FT_STATUS (CAL_CONV *pfunc_FT_GetDeviceInfoList)(
		FT_DEVICE_LIST_INFO_NODE *pDest, LPDWORD lpdwNumDevs);
typedef FT_STATUS (CAL_CONV *pfunc_FT_CreateDeviceInfoList)(LPDWORD lpdwNumDevs);
typedef FT_STATUS (CAL_CONV *pfunc_FT_OpenEx)(PVOID pArg1, DWORD Flags,
		FT_HANDLE *pHandle);
typedef FT_STATUS (CAL_CONV *pfunc_FT_Close)(FT_HANDLE ftHandle);
typedef FT_STATUS (CAL_CONV *pfunc_FT_ResetDevice)(FT_HANDLE ftHandle);
typedef FT_STATUS (CAL_CONV *pfunc_FT_Purge)(FT_HANDLE ftHandle, DWORD dwMask);
typedef FT_STATUS (CAL_CONV *pfunc_FT_SetUSBParameters)(FT_HANDLE ftHandle,
		DWORD dwInTransferSize, DWORD dwOutTransferSize);
typedef FT_STATUS (CAL_CONV *pfunc_FT_SetChars)(FT_HANDLE ftHandle,
		UCHAR uEventCh, UCHAR uEventChEn, UCHAR uErrorCh, UCHAR uErrorChEn);
typedef FT_STATUS (CAL_CONV *pfunc_FT_SetTimeouts)(FT_HANDLE ftHandle,
		DWORD dwReadTimeout, DWORD dwWriteTimeout);
typedef FT_STATUS (CAL_CONV *pfunc_FT_SetLatencyTimer)(FT_HANDLE ftHandle,
		UCHAR ucTimer);
typedef FT_STATUS (CAL_CONV *pfunc_FT_SetBitmode)(FT_HANDLE ftHandle,
		UCHAR ucMask, UCHAR ucMode);
typedef FT_STATUS (CAL_CONV *pfunc_FT_GetQueueStatus)(FT_HANDLE ftHandle,
		LPDWORD lpdwAmountInRxQueue);
typedef FT_STATUS (CAL_CONV *pfunc_FT_Read)(FT_HANDLE ftHandle, LPVOID lpBuffer,
		DWORD dwBytesToRead, LPDWORD lpdwBytesReturned);
typedef FT_STATUS (CAL_CONV *pfunc_FT_Write)(FT_HANDLE ftHandle,
		LPVOID lpBuffer, DWORD dwBytesToWrite, LPDWORD lpdwBytesWritten);
typedef FT_STATUS (CAL_CONV *pfunc_FT_GetDeviceInfo)(FT_HANDLE ftHandle,
		FT_DEVICE *lpftDevice, LPDWORD lpdwID, PCHAR SerialNumber,
		PCHAR Description, LPVOID Dummy);
typedef FT_STATUS (CAL_CONV *pfunc_FT_SetRts)(FT_HANDLE ftHandle);
typedef FT_STATUS (CAL_CONV *pfunc_FT_ClrDtr)(FT_HANDLE ftHandle);
typedef FT_STATUS (CAL_CONV *pfunc_FT_SetBaudRate)(FT_HANDLE ftHandle,
		ULONG BaudRate);
typedef FT_STATUS (CAL_CONV *pfunc_FT_SetDataCharacteristics)(
		FT_HANDLE ftHandle, UCHAR WordLength, UCHAR StopBits, UCHAR Parity);
typedef FT_STATUS (CAL_CONV *pfunc_FT_SetFlowControl)(FT_HANDLE ftHandle,
		USHORT FlowControl, UCHAR XonChar, UCHAR XoffChar);

pfunc_FT_GetLibraryVersion pFT_GetLibraryVersion;
pfunc_FT_GetNumChannel pFT_GetNumChannel;
pfunc_FT_GetDeviceInfoList pFT_GetDeviceInfoList;
pfunc_FT_CreateDeviceInfoList pFT_CreateDeviceInfoList;
pfunc_FT_OpenEx pFT_OpenEx;
pfunc_FT_Close pFT_Close;
pfunc_FT_ResetDevice pFT_ResetDevice;
pfunc_FT_Purge pFT_Purge;
pfunc_FT_SetUSBParameters pFT_SetUSBParameters;
pfunc_FT_SetChars pFT_SetChars;
pfunc_FT_SetTimeouts pFT_SetTimeouts;
pfunc_FT_SetLatencyTimer pFT_SetLatencyTimer;
pfunc_FT_SetBitmode pFT_SetBitmode;
pfunc_FT_GetQueueStatus pFT_GetQueueStatus;
pfunc_FT_Read pFT_Read;
pfunc_FT_Write pFT_Write;
pfunc_FT_GetDeviceInfo pFT_GetDeviceInfo;
pfunc_FT_SetRts pFT_SetRts;
pfunc_FT_ClrDtr pFT_ClrDtr;
pfunc_FT_SetBaudRate pFT_SetBaudRate;
pfunc_FT_SetDataCharacteristics pFT_SetDataCharacteristics;
pfunc_FT_SetFlowControl pFT_SetFlowControl;

// stdin monitoring thread.
#ifdef __WIN32
long hMonitorThread;
HANDLE inMutex;
#else
pthread_t hMonitorThread;
pthread_mutex_t inMutex = PTHREAD_MUTEX_INITIALIZER;
#endif
uint8_t finInBuf[512];
volatile size_t fileInLen;
volatile int fileInEof = 0;

/* LOCAL FUNCTIONS / INLINES *******************************************************/

int ftPortconfigure(void);

const char * ftStatusString(FT_STATUS status)
{
	switch (status)
	{
	case FT_OK:
		return "OK";
	case FT_INVALID_HANDLE:
		return "INVALID_HANDLE";
	case FT_DEVICE_NOT_FOUND:
		return "DEVICE_NOT_FOUND";
	case FT_DEVICE_NOT_OPENED:
		return "DEVICE_NOT_OPENED";
	case FT_IO_ERROR:
		return "IO_ERROR";
	case FT_INSUFFICIENT_RESOURCES:
		return "INSUFFICIENT_RESOURCES";
	case FT_INVALID_PARAMETER:
		return "INVALID_PARAMETER";
	case FT_INVALID_BAUD_RATE:
		return "INVALID_BAUD_RATE";
	case FT_DEVICE_NOT_OPENED_FOR_ERASE:
		return "DEVICE_NOT_OPENED_FOR_ERASE";
	case FT_DEVICE_NOT_OPENED_FOR_WRITE:
		return "DEVICE_NOT_OPENED_FOR_WRITE";
	case FT_FAILED_TO_WRITE_DEVICE:
		return "FAILED_TO_WRITE_DEVICE";
	case FT_EEPROM_READ_FAILED:
		return "EEPROM_READ_FAILED";
	case FT_EEPROM_WRITE_FAILED:
		return "EEPROM_WRITE_FAILED";
	case FT_EEPROM_ERASE_FAILED:
		return "EEPROM_ERASE_FAILED";
	case FT_EEPROM_NOT_PRESENT:
		return "EEPROM_NOT_PRESENT";
	case FT_EEPROM_NOT_PROGRAMMED:
		return "EEPROM_NOT_PROGRAMMED";
	case FT_INVALID_ARGS:
		return "INVALID_ARGS";
	case FT_NOT_SUPPORTED:
		return "NOT_SUPPORTED";
	case FT_OTHER_ERROR:
		return "OTHER_ERROR";
	case FT_DEVICE_LIST_NOT_READY:
		return "DEVICE_LIST_NOT_READY";
	default:
		return "unknown error";
	}
}

int ftPortOpen(char *portName, uint32_t openBy)
{
	FT_DEVICE_LIST_INFO_NODE *devInfo;
	DWORD numDevs;
	FT_STATUS ftStatus = FT_OK;
	int retVal = TTYUSB_NORMAL;
	int i;
	char *dportname = NULL;

	if (hdll_d2xx == NULL)
	{
#ifdef __WIN32
		if (!(hdll_d2xx = LoadLibrary("ftd2xx.dll")))
		{
			fprintf(stderr, "Error loading \"ftd2xx.dll\".\r\n");
			return TTYUSB_EXIT_NO_D2XX;
		}
#else
		hdll_d2xx = (HANDLE)1; // Non-zero
#endif
		pFT_GetLibraryVersion = (pfunc_FT_GetLibraryVersion) GET_FUNC(hdll_d2xx,
				"FT_GetLibraryVersion");
		pFT_GetNumChannel = (pfunc_FT_GetNumChannel) GET_FUNC(hdll_d2xx,
				"FT_CreateDeviceInfoList");
		pFT_GetDeviceInfoList = (pfunc_FT_GetDeviceInfoList) GET_FUNC(hdll_d2xx,
				"FT_GetDeviceInfoList");
		pFT_CreateDeviceInfoList = (pfunc_FT_CreateDeviceInfoList) GET_FUNC(
				hdll_d2xx, "FT_CreateDeviceInfoList");
		pFT_OpenEx = (pfunc_FT_OpenEx) GET_FUNC(hdll_d2xx, "FT_OpenEx");
		pFT_Close = (pfunc_FT_Close) GET_FUNC(hdll_d2xx, "FT_Close");
		pFT_ResetDevice = (pfunc_FT_ResetDevice) GET_FUNC(hdll_d2xx,
				"FT_ResetDevice");
		pFT_Purge = (pfunc_FT_Purge) GET_FUNC(hdll_d2xx, "FT_Purge");
		pFT_SetUSBParameters = (pfunc_FT_SetUSBParameters) GET_FUNC(hdll_d2xx,
				"FT_SetUSBParameters");
		pFT_SetChars = (pfunc_FT_SetChars) GET_FUNC(hdll_d2xx, "FT_SetChars");
		pFT_SetTimeouts = (pfunc_FT_SetTimeouts) GET_FUNC(hdll_d2xx,
				"FT_SetTimeouts");
		pFT_SetLatencyTimer = (pfunc_FT_SetLatencyTimer) GET_FUNC(hdll_d2xx,
				"FT_SetLatencyTimer");
		pFT_SetBitmode = (pfunc_FT_SetBitmode) GET_FUNC(hdll_d2xx,
				"FT_SetBitMode");
		pFT_GetQueueStatus = (pfunc_FT_GetQueueStatus) GET_FUNC(hdll_d2xx,
				"FT_GetQueueStatus");
		pFT_Read = (pfunc_FT_Read) GET_FUNC(hdll_d2xx, "FT_Read");
		pFT_Write = (pfunc_FT_Write) GET_FUNC(hdll_d2xx, "FT_Write");
		pFT_GetDeviceInfo = (pfunc_FT_GetDeviceInfo) GET_FUNC(hdll_d2xx,
				"FT_GetDeviceInfo");
		pFT_SetRts = (pfunc_FT_SetRts) GET_FUNC(hdll_d2xx, "FT_SetRts");
		pFT_ClrDtr = (pfunc_FT_ClrDtr) GET_FUNC(hdll_d2xx, "FT_ClrDtr");
		pFT_SetBaudRate = (pfunc_FT_SetBaudRate) GET_FUNC(hdll_d2xx,
				"FT_SetBaudRate");
		pFT_SetDataCharacteristics = (pfunc_FT_SetDataCharacteristics) GET_FUNC(
				hdll_d2xx, "FT_SetDataCharacteristics");
		pFT_SetFlowControl = (pfunc_FT_SetFlowControl) GET_FUNC(hdll_d2xx,
				"FT_SetFlowControl");
	}

	// Try to open the supplied port
	if (strlen(portName))
	{
		ftStatus = pFT_OpenEx(portName, openBy, &ftInterface);

		if (ftStatus != FT_OK)
		{
			fprintf(stderr, "Error: Failed to open port: %s\r\n", portName);
			retVal = TTYUSB_ERROR;
			goto exit;
		}

		retVal = ftPortconfigure();
		dportname = strdup(portName);
	}
	else
	{
		// find available devices here
		ftStatus = pFT_CreateDeviceInfoList(&numDevs);
		if (ftStatus != FT_OK)
		{
			fprintf(stderr,
					"FT_CreateDeviceInfoList failed (%s).\r\n",
					ftStatusString(ftStatus));
			retVal = TTYUSB_ERROR;
			goto exit;
		}

		if (numDevs == 0)
		{
			fprintf(stdout, "No devices available.\r\n");
			retVal = TTYUSB_ERROR;
		}
		else
		{
			// if there are devices connected, display their names
			// allocate storage for our list
			devInfo = (FT_DEVICE_LIST_INFO_NODE*) malloc(
					sizeof(FT_DEVICE_LIST_INFO_NODE) * numDevs);
			// populate the device information list
			ftStatus = pFT_GetDeviceInfoList(devInfo, &numDevs);
			if (ftStatus != FT_OK)
			{
				fprintf(stderr,
						"FT_GetDeviceInfoList failed (%s).\r\n",
						ftStatusString(ftStatus));
				retVal = TTYUSB_ERROR;
				free(devInfo);
				goto exit;
			}

			fprintf(stdout, "Available devices are:\r\n");
			for (i = 0; i < (int) numDevs; i++)
			{
				// print the device description
				fprintf(stdout, "Description: \"%s\". Serial Number: %s.\r\n",
						devInfo[i].Description, devInfo[i].SerialNumber);

				ftStatus = pFT_OpenEx((PVOID)devInfo[i].Description, FT_OPEN_BY_DESCRIPTION,
						&ftInterface);

				if (ftStatus == FT_OK)
				{
					if (listAll == 0)
					{
						dportname = strdup(portName);
						break;
					}
					pFT_Close(ftInterface);
				}
				else if (i == (int)numDevs - 1)
				{
					// Failed to open any port
					fprintf(stderr,
							"Failed to open any port.\r\n"
							"Most recent FT_OpenEx error was %s.\r\n",
							ftStatusString(ftStatus));
				}
			}
			free(devInfo);
		}
	}

	exit:
	if (retVal == TTYUSB_ERROR)
	{
		if (dportname)
		{
			free(dportname);
		}
		return TTYUSB_ERROR;
	}

	if (logOutput)
	{
		fprintf(stdout, "Opened D2XX port: %s\r\n", dportname);
	}

	free(dportname);
	return retVal;
}

int ftPortClose(void)
{
	if (ftInterface)
	{
		pFT_Close(ftInterface);
		ftInterface = NULL;
	}

	if (hdll_d2xx)
	{
#ifdef __WIN32
		FreeModule(hdll_d2xx);
#endif
	}

	return TTYUSB_NORMAL;
}

int ftPortconfigure(void)
{
	FT_STATUS ftStatus;

	ftStatus = pFT_Purge(ftInterface, FT_PURGE_RX | FT_PURGE_TX);
	if (ftStatus != FT_OK)
	{
		fprintf(stderr, "FT_Purge failed (%s).\r\n", ftStatusString(ftStatus));
		return TTYUSB_ERROR;
	}

	ftStatus = pFT_SetTimeouts(ftInterface, 200, 0);
	if (ftStatus != FT_OK)
	{
		fprintf(stderr, "FT_SetTimeouts failed (%s)\r\n",
				ftStatusString(ftStatus));
		return TTYUSB_ERROR;
	}

	ftStatus = pFT_SetLatencyTimer(ftInterface, 16);
	if (ftStatus != FT_OK)
	{
		fprintf(stderr, "FT_SetLatency failed (%s).\r\n",
				ftStatusString(ftStatus));
		return TTYUSB_ERROR;
	}

	ftStatus = pFT_SetRts(ftInterface);
	if (ftStatus != FT_OK)
	{
		fprintf(stderr, "FT_SetRts failed (%s)\r\n", ftStatusString(ftStatus));
		return TTYUSB_ERROR;
	}

	ftStatus = pFT_ClrDtr(ftInterface);
	if (ftStatus != FT_OK)
	{
		fprintf(stderr, "FT_ClrDtr failed (%s)\r\n", ftStatusString(ftStatus));
		return TTYUSB_ERROR;
	}

	ftStatus = pFT_SetBaudRate(ftInterface, 19200);
	if (ftStatus != FT_OK)
	{
		fprintf(stderr, "FT_SetBaudRate failed (%s)\r\n",
				ftStatusString(ftStatus));
		return TTYUSB_ERROR;
	}

	ftStatus = pFT_SetDataCharacteristics(ftInterface,
			FT_BITS_8,
			FT_STOP_BITS_1,
			FT_PARITY_NONE);
	if (ftStatus != FT_OK)
	{
		fprintf(stderr, "FT_SetDataCharacteristics failed (%s)\r\n",
				ftStatusString(ftStatus));
		return TTYUSB_ERROR;
	}

	ftStatus = pFT_SetFlowControl(ftInterface, FT_FLOW_NONE, 0, 0);
	if (ftStatus != FT_OK)
	{
		fprintf(stderr, "FT_SetFlowControl failed (%s)\r\n",
				ftStatusString(ftStatus));
		return TTYUSB_ERROR;
	}

	return TTYUSB_NORMAL;
}

// ---------------------------------------------
// Write the buffer to the USB device
//

int ftWrite(uint8_t *buffer, size_t len)
{
	DWORD     bytesWritten;
	FT_STATUS ftStatus;

	if (logOutput > 1)
	{
		fprintf(stdout, "Writing %d bytes\r\n", len);
	}

	ftStatus = pFT_Write(ftInterface,
			buffer,
			(DWORD)len,
			&bytesWritten);
	if (ftStatus != FT_OK)
	{
		fprintf(stderr, "FT_Write failed (%s).\r\n", ftStatusString(ftStatus));
		return TTYUSB_ERROR;
	}

	// Check that the number of bytes written makes sense.
	if ((int)bytesWritten != len)
	{
		fprintf(stderr,
				"FT_Write failed: wrote %d of %d bytes.\r\n",
				(int)bytesWritten,
				len);
		return TTYUSB_ERROR;
	}

	return TTYUSB_NORMAL;
}

// ---------------------------------------------
// Read data to a buffer from the USB device
//
int ftRead(uint8_t *buffer, size_t *len)
{
	FT_STATUS ftStatus;
	DWORD bytesRead;
	DWORD bytesWaiting;

	ftStatus = pFT_GetQueueStatus(ftInterface, &bytesWaiting);
	if (ftStatus != FT_OK)
	{
		fprintf(stderr, "FT_Read/FT_GetQueueStatus failed (%s).\r\n", ftStatusString(ftStatus));
		return TTYUSB_ERROR;
	}

	if (bytesWaiting > 0)
	{
		ftStatus = pFT_Read(ftInterface,
				buffer,
				bytesWaiting,
				&bytesRead);

		if (ftStatus != FT_OK)
		{
			fprintf(stderr, "FT_Read failed (%s).\r\n", ftStatusString(ftStatus));
			return TTYUSB_ERROR;
		}

		*len = (int)bytesRead;
	}
	else
	{
		*len = 0;
	}

	return TTYUSB_NORMAL;
}

/**
    @brief      Print version information
 **/
void printVersionInfo(void)
{
	printf("V" TTYUSB_VERSION " (C) Copyright 2017, Bridgetek Pte. Ltd.\r\n");
}

/**
    @brief      Print usage information
 **/
static void usage(char *exeName)
{
	const char *args =
			"  -{h|?}, --help  Display this help\r\n"
			"  -a              List available devices\r\n"
			"  -p <device>     Use device name\r\n"
			"  -o <filename>   Output filename (- for stdout)\r\n"
			"  -f <filename>   Input filename (- for stdin)\r\n"
			"  -m              Monitor input file for changes\r\n"
			"Neither -o nor -f means both stdin/stdout\r\n"
			"\r\n";

	printf("\nUsage: %s [args]\n\r\n",
			exeName);
	printf(args);
	printVersionInfo();
}

/**
    @brief      Parse the command line for options
    @details	Step through the command line parameters to find
    			options and input files.
 **/
static int parseCmdLine (int argc, char **argv)
{
	int i;
	int status = TTYUSB_NORMAL;

	// defaults
	inFilename = NULL;
	outFilename = NULL;
	useStdin = 0;
	useStdout = 0;
	useStdinout = 1;
	logOutput = 0;

	for (i = 1; i < argc ; i++)
	{
		/* if this is an option */
		if (argv[i][0] == '-')
		{
			/* display usage */
			if (strcmp(argv[i], "-h") == 0 ||
					strcmp(argv[i], "-?") == 0 ||
					strcmp(argv[i], "--help") == 0)
			{
				usage(argv[0]);
				exit(TTYUSB_NORMAL);
			}

			/* display version */
			if (strcmp(argv[i], "--version") == 0)
			{
				printVersionInfo();
				exit(TTYUSB_NORMAL);
			}

			/* show logging information */
			if (strcmp(argv[i], "-l") == 0)
			{
				logOutput = 1;
				continue;
			}
			if (strcmp(argv[i], "-la") == 0)
			{
				logOutput = 2;
				continue;
			}

			if (strcmp(argv[i], "-m") == 0)
			{
				monitoring = 1;
				continue;
			}

			if (strcmp(argv[i], "-p") == 0)
			{
				if (listAll > 0)
				{
					fprintf(stderr, "ERROR: -p and -a cannot both be used\r\n");
					status = TTYUSB_EXIT_PARAMETERS;
					break;
				}
				if (i+1 >= argc)
				{
					fprintf(stderr, "ERROR: -p needs D2XX device name\r\n");
					status = TTYUSB_EXIT_PARAMETERS;
					break;
				}
				i++;
				ttyUSBPortName = argv[i];
				continue;
			}

			if (strcmp(argv[i], "-a") == 0)
			{
				if (strlen(ttyUSBPortName) > 0)
				{
					fprintf(stderr, "ERROR: -p and -a cannot both be used\r\n");
					status = TTYUSB_EXIT_PARAMETERS;
					break;
				}
				listAll = 1;
				continue;
			}

			if (strcmp(argv[i], "-o") == 0)
			{
				useStdinout = 0;
				i++;
				if (strcmp(argv[i], "-") == 0)
				{
					useStdout = 1;
				}
				else
				{
					outFilename = argv[i];
					useStdout = 0;
				}
				continue;
			}

			if (strcmp(argv[i], "-f") == 0)
			{
				useStdinout = 0;
				i++;
				if (strcmp(argv[i], "-") == 0)
				{
					useStdin = 1;
				}
				else
				{
					inFilename = argv[i];
					useStdin = 0;
				}
				continue;
			}

			fprintf(stderr, "ERROR: unknown option %s --- ignored\r\n", argv[i]);
		}
	}

	return status;
}

#ifdef __WIN32
void fileMonitor(void *param)
{
	uint8_t inBuf[sizeof(finInBuf)];
	char *paramInFilePath = (char *)param;
	HANDLE hFileHandle;
	DWORD rdLen = 0;
	int ending = 0;
	int changed = 1;
	FILE_NOTIFY_INFORMATION strFileNotifyInfo[1024];
	DWORD dwBytesReturned = 0;
	wchar_t *wInFilePath;
	wchar_t *wInFullPath;
	wchar_t *wInPath;
	wchar_t *wInFullPathFilePart;
	wchar_t *wInFileName;
	int i;
	HANDLE hDir;

	// Convert char* string to a wchar_t* string.
	wInFilePath = (wchar_t *)malloc((strlen(paramInFilePath) + 1) * 2);
	mbstowcs(wInFilePath, paramInFilePath, (strlen(paramInFilePath) + 1) * 2);

	dwBytesReturned = GetFullPathNameW(wInFilePath, 0, NULL, NULL);
	wInFullPath = (wchar_t *)malloc((dwBytesReturned + 1) * 2);
    GetFullPathNameW(wInFilePath, dwBytesReturned, wInFullPath, &wInFullPathFilePart);

    wInFileName = (wchar_t *)malloc((wcslen(wInFullPathFilePart) + 1) * 2);
    wcscpy(wInFileName, wInFullPathFilePart);
	for (i = 0; i < wcslen(wInFileName); i++)
	{
		wInFileName[i] = towlower(wInFileName[i]);
	}

    *wInFullPathFilePart = 0;
    wInPath = (wchar_t *)malloc((wcslen(wInFullPath) + 1) * 2);
    wcscpy(wInPath, wInFullPath);
    free(wInFullPath);

    SetCurrentDirectoryW(wInPath);
    hDir = CreateFileW(
    		wInPath,
			FILE_LIST_DIRECTORY,
			FILE_SHARE_WRITE | FILE_SHARE_READ | FILE_SHARE_DELETE,
			NULL,
			OPEN_EXISTING,
			FILE_FLAG_BACKUP_SEMANTICS,
			NULL);

	do
	{
		if (changed == 0)
		{
			if (ReadDirectoryChangesW(hDir,
					(LPVOID)&strFileNotifyInfo, sizeof(strFileNotifyInfo),
					FALSE, FILE_NOTIFY_CHANGE_LAST_WRITE, &dwBytesReturned, NULL, NULL) == 0)
			{
				fprintf(stderr, "Reading Directory Change error\r\n");
				break;
			}
			else
			{
				if (logOutput > 1)
				{
					printf("File changed\r\n");
				}

				for (i = 0; i < wcslen(strFileNotifyInfo[0].FileName); i++)
				{
					strFileNotifyInfo[0].FileName[i] = towlower(strFileNotifyInfo[0].FileName[i]);
				}

				if ((wcscmp(strFileNotifyInfo[0].FileName, wInFileName) == 0) || (ending == 0))
				{
					if (logOutput > 1)
					{
						printf("Target file \"%s\" changed\r\n", paramInFilePath);
					}
					changed = 1;
				}
			}
		}

		if (changed)
		{
			rdLen = 0;

			hFileHandle = CreateFile(inFilename,
					GENERIC_READ, FILE_SHARE_WRITE | FILE_SHARE_READ | FILE_SHARE_DELETE,
					NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

			if (hFileHandle == INVALID_HANDLE_VALUE)
			{
				fprintf(stderr, "Could not open file \"%s\"\r\n", paramInFilePath);
				hFileHandle = NULL;
			}
			else
			{
				ending = 0;

				while (!ending)
				{
					if (rdLen == 0)
					{
						ending = !ReadFile(hFileHandle,
								inBuf, sizeof(finInBuf),
								&rdLen, NULL);
					}

#ifdef __WIN32
					if (WaitForSingleObject(inMutex, 0) == WAIT_OBJECT_0)
#else
					if (pthread_mutex_lock(&inMutex) == 0)
#endif
					{
						if (fileInLen == 0)
						{
							if ((rdLen > 0) && (!ending))
							{
								fileInLen = (size_t)rdLen;
								memcpy(finInBuf, inBuf, rdLen);
							}
							else
							{
								ending = 1;
							}

							rdLen = 0;
						}
#ifdef __WIN32
						ReleaseMutex(inMutex);
#else
						pthread_mutex_unlock(&inMutex);
#endif
					}

					changed = 0;
				}

				if (logOutput > 1)
				{
					printf("Closing file\r\n");
				}
				CloseHandle(hFileHandle);
			}
		}
	} while (monitoring);

	CloseHandle(hDir);
	free(wInFilePath);
	free(wInFileName);
	free(wInPath);

	fileInEof = EOF;
	if (logOutput)
	{
		printf("Closing file thread\r\n");
	}
}
#else
void fileMonitor(void *param)
{
	uint8_t inBuf[sizeof(finInBuf)];
	char *paramInFilePath = (char *)param;
	HANDLE hFileHandle;

	printf("FILENAME: %s\r\n", paramInFilePath);
}
#endif // __WIN32

void stdinMonitor(void *param)
{
	uint8_t inBuf[sizeof(finInBuf)];
#ifdef __WIN32
	HANDLE hStdin = (HANDLE)param;
	DWORD rdLen = 0;
#else
	FILE *hStdin = (FILE *)param;
	size_t rdLen = 0;
#endif
	int ending = 0;

	while (!ending)
	{
#ifdef __WIN32

		if (rdLen == 0)
		{
			ending = !ReadFile(hStdin,
					inBuf, sizeof(finInBuf),
					&rdLen, NULL);
		}

#ifdef __WIN32
		if (WaitForSingleObject(inMutex, 0) == WAIT_OBJECT_0)
#else
		if (pthread_mutex_lock(&inMutex) == 0)
#endif
		{
			if (fileInLen == 0)
			{
				if ((rdLen > 0) && (!ending))
				{
					fileInLen = (size_t)rdLen;
					memcpy(finInBuf, inBuf, rdLen);
				}
				else
				{
					fileInEof = EOF;
					ending = 1;
				}

				rdLen = 0;
			}
#ifdef __WIN32
			ReleaseMutex(inMutex);
#else
			pthread_mutex_unlock(&inMutex);
#endif
		}
#else
#if 0
		kbB = kbhit();
		if (kbB)
		{
			readB = getch();
			printf(")");
			if ((readB == EOF)
					|| (readB == 0x03))
			{
				break;
			}
		}
#endif
#endif
	}

	if (logOutput)
	{
		printf("Closing stdin thread\r\n");
	}
}

int main(int argc, char **argv)
{
	int status = 0;
	uint8_t inBuf[512];
	uint8_t outBuf[512];
	size_t inLen;
	size_t outLen;
	FT_STATUS ftret;
	int inEof;
#ifdef __WIN32
	HANDLE fileIn = NULL;
	HANDLE fileOut = NULL;
	DWORD cWritten, fdwMode, fdwOldMode;
#else
	FILE *fileIn = NULL;
	FILE *fileOut = NULL;
#endif

#ifdef __WIN32
	inMutex = CreateMutex(NULL, FALSE, NULL);
#else
	pthread_mutex_init(&inMutex, NULL);
#endif

	ttyUSBPortName = "";
	status = parseCmdLine(argc, argv);

	printf("Starting\r\n");

	if (status == 0)
	{
		if ((ttyUSBPortName) || (listAll))
		{
			if (ftPortOpen(ttyUSBPortName, FT_OPEN_BY_DESCRIPTION) != 0)
			{
				fprintf(stderr, "Cannot open port \"%s\"\r\n", ttyUSBPortName);
				return TTYUSB_EXIT_NO_D2XX;
			}

			fileInLen = 0;
			fileInEof = 0;
			inLen = 0;
			inEof = 0;

			if (!listAll)
			{
				if ((useStdin) || (useStdinout))
				{
					if (logOutput)
					{
						fprintf(stdout, "Opening \"stdin\"\r\n");
					}
#if __WIN32
					fileIn = GetStdHandle(STD_INPUT_HANDLE);
					if (fileIn == INVALID_HANDLE_VALUE)
					{
						fprintf(stderr, "GetStdHandle failed\r\n");
						return 1;
					}

					// Turn off the line input and echo input modes
					GetConsoleMode(fileIn, &fdwOldMode);
					fdwMode = fdwOldMode & ~(ENABLE_LINE_INPUT | ENABLE_ECHO_INPUT);
					SetConsoleMode(fileIn, fdwMode);

					hMonitorThread = _beginthreadex(NULL, 0, (void *)stdinMonitor, (void *)fileIn, 0, NULL);
					if (hMonitorThread == (long)NULL)
					{
						fprintf(stdout, "No monitor thread\r\n");
					}
#else
					fileIn = stdin;
					if (pthread_create(&hMonitorThread, NULL, (void *)stdinMonitor, (void *)fileIn) != 0)
					{
						fprintf(stdout, "No monitor thread\r\n");
					}
#endif
				}
				else if (inFilename)
				{
					if (logOutput)
					{
						fprintf(stdout, "Opening in file \"%s\"\r\n", inFilename);
					}
#ifdef __WIN32
					hMonitorThread = _beginthreadex(NULL, 0, (void *)fileMonitor, (void *)inFilename, 0, NULL);
					if (hMonitorThread == (long)NULL)
					{
						fprintf(stdout, "No monitor thread\r\n");
					}
#else
					if (pthread_create(&hMonitorThread, NULL, (void *)fileMonitor, (void *)inFilename) != 0)
					{
						fprintf(stdout, "No monitor thread\r\n");
					}
#endif
				}

				if ((useStdout) || (useStdinout))
				{
					if (logOutput)
					{
						fprintf(stdout, "Opening \"stdout\"\r\n");
					}
#if __WIN32
					fileOut = GetStdHandle(STD_OUTPUT_HANDLE);
					if (fileOut == INVALID_HANDLE_VALUE)
					{
						fprintf(stderr, "GetStdHandle failed\r\n");
						return 1;
					}
#else
					fileOut = stdout;
					setmode(fileno(stdout), O_BINARY);
#endif
				}
				else if (outFilename)
				{
					if (logOutput)
					{
						fprintf(stdout, "Opening out file \"%s\"\r\n", outFilename);
					}
					fileOut = fopen(outFilename, "wb");
					if (fileOut == NULL)
					{
						fprintf(stderr, "Could not open out file \"%s\"\r\n", outFilename);
					}
				}

				while (1)
				{
#ifdef __WIN32
					if (WaitForSingleObject(inMutex, 5000) == WAIT_OBJECT_0)
					{
						inEof = fileInEof;

						if (fileInLen > 0)
						{
							inLen = fileInLen;
							if (inLen > sizeof(inBuf))
							{
								inLen = sizeof(inBuf);
							}
							memcpy(inBuf, finInBuf, inLen);
							fileInLen = 0;
						}
						else
						{
							inLen = 0;
						}
						ReleaseMutex(inMutex);
					}
#else
					{
						inLen = fread(inBuf, 1, sizeof(inBuf), fileIn);
						inEof = (inLen < sizeof(inBuf))?EOF:0;
					}
#endif
					if (inLen > 0)
					{
						ftret = ftWrite(inBuf, inLen);
						if (logOutput || (ftret != FT_OK))
						{
							fprintf(stderr, "FT_STATUS ftWrite %d\r\n", (int)ftret);
						}
						if (ftret != FT_OK)
						{
							break;
						}
					}

					outLen = sizeof(outBuf);
					ftret = ftRead(outBuf, &outLen);
					if (ftret == FT_OK)
					{
						if (outLen > 0)
						{
#ifndef __WIN32
							outLen = fwrite(outBuf, 1, outLen, fileOut);
#else
							if (!WriteFile(fileOut,
									outBuf, outLen, &cWritten, NULL))
							{
								printf("WriteFile\r\n");
							}
#endif
							if (logOutput || (outLen < 0))
							{
								fprintf(stderr, "fwrite %d\r\n", (int)outLen);
							}
							if (outLen < 0)
							{
								break;
							}
						}
					}
					else
					{
						fprintf(stderr, "FT_STATUS ftRead %d\r\n", (int)ftret);
						break;
					}

					if (!useStdinout)
					{
						if (inEof == EOF)
						{
							break;
						}
					}
				} // while

				if (logOutput || (outLen < 0))
				{
					printf("Exiting\r\n");
				}

#ifdef __WIN32
				if (fileIn)
					CloseHandle(fileIn);
				if (fileOut)
					CloseHandle(fileOut);
#else
				if (fileIn)
					fclose(fileIn);
				if (fileOut)
					fclose(fileOut);
#endif
			}
			ftPortClose();
		}
	}
	else
	{
		usage(argv[0]);
	}

#ifdef __WIN32
	CloseHandle(inMutex);
#else
	pthread_mutex_destroy(&inMutex);
#endif

	return status;
}
